TITLE Terminal Emulation for use with asynchronous device driver.
NAME ATERM
PAGE 70,132 ;set the page length, width
;************************************************************************
;*									*
;*	      Aterm- a simple terminal emulation program for		*
;*	      =====  use with MS-DOS and Brain Boxes Asynchronous	*
;*		     Device Driver.					*
;*									*
;*									*
;*		    written by Richard Davis 10/5/1990			*
;*		      Copyright (c) 1990 Brain Boxes			*
;*									*
;*									*
;************************************************************************
;
SUBTTL MS-DOS Routine Equates
PAGE
;
key = 016h ;rom bios keyboard services
DOS = 21H ;DOS Function Call put function in .ah
video = 10H ;rom bios video light pen interrupt is service 4
PRT = 17H ;rom bios printer interrupt
COM = 14H ;rom bios serial port interrupt
;
seti =025h ;set interrupt vector
geti =035h ;get interrupt vector
cr = 0dh ;carriage return
lf = 0ah ;line feed
newl = lf			;Internal newline character.
nul equ 0
spce = 020h ;space
tab equ 9			;Tab
shriek = 021h ;!
quotes1 = 027h ;single quotes
quotes2 = 022h ;double quotes
comma = 02ch ;comma
colon equ 03AH			;Colon
bel equ 7			;Bell (bleep)
;
NULL equ 0			;NULL pointer
;
tolower equ 00100000b		;OR mask for upper to lower case conversion
toupper equ 11011111b		;AND mask for lower to upper case conversion
;
;
SUBTTL Stack definition
PAGE
;
stack segment word stack 'stack'
;every program needs a stack
 db 1024 dup (0)		;1024 bytes of stack space
 stack_size equ $		;length of stack segment
stack ends
;
;
SUBTTL Global Data
PAGE
;
; Define Data segment
;
data segment word public
;
;Define serial port- Numbers are as used by BIOS extended initialise.
  comno    dw 0 		;COM1 is default (0 is used by BIOS services)
  comname  db 'COM1',0,0,0,0,0  ;Nul-terminated name, up to 8 characters
  baud	   db 4 		; ==> 1200 Baud
  parity   db 2 		; ==> even parity
  databits db 2 		; ==> 7 data bits
  stopbits db 0 		; ==> 1 stop bit
  brk	   db 0 		; ==> no break
;
  dabbasechar equ '5'           ;databits number = 0 ==> '5' data bits
  stobasechar equ '1'           ;stopbits number = 0 ==> '1' stop bit
;
  align 2
  comport dw -1 		; File handle for com port.
;
;Define Terminal operation.
  echon EQU 0			; No echo
  echos EQU 1			; Echo keyboard to screen
  echol EQU 2			; Echo from port back to port
  echols EQU 3			; Echo both
;
  echo dw echon 		; No echo
;
; Translation tables. Each is list of pointers to list of substitution
; Characters.
  xlt_size EQU 1024
;
IF1				; Macro are a little slow to assemble
				; Use only on first pass.
;
xlt_sub MACRO size		; Macro defines a substitution table
  local i
  i = 0
  REPT size
  db i				; Character i translates to i
  i = (i + 1) AND 255
  ENDM				; End repeat
  ENDM				; End xlt_tbl
;
xlt_tbl MACRO sub_addr		; Macro defines a translation table
  local j
  j = offset sub_addr
  REPT 257
  dw j
  j = j + 1			; Pointer to successive susbtitutions
  ENDM				; End Repeat
  ENDM				; End xlt_tbl
;
ENDIF				; Conditional assembly of macro definition
;
  .SALL 			; Suppress listing of macro expansion
;
inxltsub label byte
  xlt_sub 256			; Define input substitution
  db xlt_size - 256 dup(0)	; Leave space for substitution to grow
inxlt label word
  xlt_tbl inxltsub		; Define corresponding translation table
;
outxltsub label byte
  xlt_sub 256			; Define output substitution
  db xlt_size - 256 dup(0)	; Leave space for substitution to grow
outxlt label word
  xlt_tbl outxltsub		; Define corresponding translation table
;
;
;Data buffers
  xiolen equ 128		; Length of buffer
  xiohead2 db ? 		; 1 byte header +
  xiohead db 2 dup(?)		;   2 header bytes for DOS i/o to use
				; (Cliout, exec_dos)
  xiobuf db xiolen dup(0)	; Buffer for translation of i/o, analysis
				; of commands
;
  setxbuf db xiolen dup(?)	; Space to build translations when set
;
;temporary buffer for use by output routines
  scrtmp db 20 dup(?)
;
;Extended key codes and associated routines
cmdkeytable db 68	;F10  ==> quit
	    db 59	;F1   ==> help
	    db 60	;F2   ==> setup
	    db 46	;CTRL-C ==> Enter command
;
  numcmdkey equ 4		;Number command keys impemented.
;
cmdkeyproc dw quit		;
	   dw help		;
	   dw setup		;
	   dw cliout		;
;
;Command letters and associated routines (note letters must be lower case)
cmdtable db 'h'         ;h ==> help
	 db 'q'         ;q ==> quit
	 db 'b'         ;b ==> set Baud rate
	 db 'o'         ;i ==> set input translation
	 db 'i'         ;o ==> set output translation
	 db 'p'         ;p ==> set parity
	 db 'd'         ;d ==> set number data bits
	 db 's'         ;d ==> set number stop bits
	 db 'e'         ;e ==> set echo
	 db 'c'         ;c ==> set com port
	 db 't'         ;t ==> transmit sequence
	 db quotes1	;' ==> comment
	 db 'x'         ;x ==> execute file
	 db 'r'         ;r ==> execute DOS command
;
  numcmd equ 14 		;Number commands
;
cmdproc dw help 		;
	dw quit 		;
	dw set_baud		;
	dw set_outxlt		;
	dw set_inxlt		;
	dw set_parity		;
	dw set_databits 	;
	dw set_stopbits 	;
	dw set_echo		;
	dw set_com		;
	dw transmit		;
	dw send_comment 	;
	dw exec_file		;
	dw exec_dos		;
;
;Prefer_baud table of baud rate truncations. Used to expand and check baud
;rate specified by user.
  pfbtable dw 11,110,0				;110 Baud
	   dw 15,150,0				;150 Baud
	   dw 30,300,0				;300 Baud
	   dw 60,600,0				;600 Baud
	   dw 12,120,1200,0			;1200 Baud
	   dw 24,240,2400,0			;2400 Baud
	   dw 48,480,4800,0			;4800 Baud
	   dw 96,960,9600,0			;9600 Baud
	   dw 19,192,1920,19200,0		;19200 Baud
	   dw 38,384,3840,38400,0		;38400 Baud
	   dw 56,560,5600,56000
	   dw 57,576,5760,57600,0		;57600 Baud
	   dw 115,1152,11520,0			;115200 Baud
	   dw 0 				;end of list
;
;Data space local to setxltlow procedure
  sxleox  dw ?
  sxldlen dw ?
  sxlxlt  dw ?
;
;Data space for set_xlt procedure
  sxtkeyst dw ?
  sxttrans dw ?
;
;data for set_parity procedure
paritychars db 'smeon'                  ;Letters associated with parity
numparych equ 5 			;(Note are in reverse order)
;
;Definition file dor use if none specified on command line
  def_file db 'termdef.txt',0
;
;location PSP-used by exec_dos and exec_defaults
  psp dw ?
;
;Parameter block and command.com pathname for use by exec_dos
  comspecname db 'COMSPEC=',0
;
  exdparams  dw 0			;environment
  exdcmdline dw ?			;offset of command line
	     dw data			;segment of command line
	     dw 05CH			;offset FCB
  exdpsp1    dw ?			;segment FCB
	     dw 06CH			;offset FCB
  exdpsp2    dw ?			;segment FCB
;
;space for menus to be built
mainspace   db 180 dup(?)
mainptspace dw 8 dup(?)
;
SUBTTL Text messages
PAGE
;
;All are nul-terminated ASCII (ASCIIZ) strings
;
openfail1 db 'Cannot open device ',0
openfail2 db '. Try a differnet name.',newl,0
;
readfail  db 'Failed to read from port.',newl,0
;
writefail db 'Failed to output to port.',newl,0
;
quitmes   db 'Exitting program.',newl,0
;
cliprompt db newl,'Type command:',0
;
exdfm1	  db 'Could not execute command.com.',newl,0
exdfm2	  db 'Could not adjust memory.',newl,0
exdfm3	  db 'Could not find path for command.com.',newl,0
;
exdm1	  db 'To end DOS shell and return to ATERM use command "exit".',newl,0
;
exffm1	  db 'No path name to execute.',newl,0
exffm2	  db 'Two many names to execute.',newl,0
exffm3	  db 'Could not execute file ',0
;
exfm1	  db 'Executing ',0
;
tnsm1	  db 'Sent to serial port sequence: ',0
;
tnsfm1	  db 'Could not understand transmit sequence.',newl,0
tnsfm2	  db 'Must be only one transmit sequence.',newl,0
;
secwm1	  db 'Device name can only be eight characters long. '
	  db 'Extra ignored.',newl,0
secwm2	  db 'Extra parameters after com port name ignored.',newl,0
;
secfm1	  db 'No name specified for serial port.',newl,0
;
secm1	  db 'Setting serial port to ',0
;
seefm1	  db 'Unrecognised character in echo specification ignored.',newl,0
;
seem1	  db 'Setting echo to ',0
;
echonames db 'no echo.',0
	  db 'echo to screen only.',0
	  db 'echo to line only.',0
	  db 'echo to screen and line.',0
;
spyf1	  db 'Cannot understand parity specification.',newl,0
spyf2	  db 'Cannot set parity.',newl,0
;
spym1	  db 'Setting parity to ',0
spym2	  db '.',newl,0
;
parynames db 'none',0
	  db 'odd',0
	  db 'even',0
	  db 'mark',0
	  db 'space',0
;
sdbmf1	  db 'Cannot understand data bits specification.',newl,0
sdbmf2	  db 'Could not set number data bits.',newl,0
;
sdbm1	  db 'Setting number data bits to ',0
;
sstmf1	  db 'Cannot understand stop bits specification.',newl,0
sstmf2	  db 'Could not set number stop bits.',newl,0
;
sstm1	  db 'Setting number stop bits to ',0
;
badcmdmes db 'Cannot understand command:-',newl,'           ',0
;
sxlfail   db 'Cannot make that translation-not enough space.',newl,0
;
badbaud   db 'Cannot set that Baud rate.',newl,0
;
baudnames db '110 Baud',0
	  db '150 Baud',0
	  db '300 Baud',0
	  db '600 Baud',0
	  db '1200 Baud',0
	  db '2400 Baud',0
	  db '4800 Baud',0
	  db '9600 Baud',0
	  db '19200 Baud',0
	  db '38400 Baud',0
	  db '57600 Baud',0
	  db '115200 Baud',0
;
setbaud1  db 'Setting Baud rate to ',0
;
sixm1	  db 'Setting input translation ',0
soxm1	  db 'Setting output translation ',0
;
sxtm1	  db 'from ',0
sxtm2	  db ' to ',0
;
sxtfm1	  db newl,'Cannot understand key character.',newl,0
sxtfm2	  db newl,'Key must be single character.',newl,0
sxtfm3	  db newl,'Key and translation must be separated by ":".',newl,0
sxtfm4	  db newl,'Cannot understand translation.',newl,0
sxtfm5	  db newl,'Extra parameters ignored.',newl,0
;
helpmes   db 'Function keys:',newl
db ' F1             help                   F10             quit',newl,newl
db 'Type ALT-C (ie press "C" key while "alt" key is held done) to issue command.',newl
db 'Commnads are:',newl
db ' H              help                   Q               quit',newl
db ' E{n|s|l|ls}    set echo screen, line  C<portname>     set port (e.g. COM2)',newl
db ' B<Baudrate>    set Baud rate          P<parity>       set parity {n|o|e|m|s}',newl
db ' D<databits>    set number data bits   S<stopbits>     set number stop bits',newl
db ' I<char>:<strn> translate input        O<char>:<strn>  translate output',newl
db ' ( <char> is a decimal number, or a character in quotation marks.',newl
db '   <strn> is sequence of <char> separated by commas, Eg',newl
db '   o12:12,',39,'New page',39,',13,10,10                                        )',newl
db ' X<filename>    execute file           R<command...>   run DOS command',newl
db spce,39,   '              comment',newl,0
;
;****************************************************************************
;Setup menus
;****************************************************************************
;
;Each menu has set of ASCIIZ options, and a list of pointers to these.
;
supm1	  db newl,newl,'Setup complete, returning to terminal session.',newl,0
;
smum1	  db newl,newl,'  Key  For....',0
smum2	  db newl,'  <',0
smum3	  db '>  ',0
smum4	  db newl,newl,' select option :',0
;
;
;Main menu
;
mainname db '==============Setup Main Menu=============',newl
	 db '=========Alter Terminal Settings==========',bel,newl,0
;
mainmn label byte
  db 'Port (now ',0,          ')',0
  db 'Baud rate (now ',0,     ')',0
  db 'Parity (now ',0,        ')',0
  db 'No. data bits (now ',0, ')',0
  db 'No. stop bits (now ',0, ')',0
  db 'Continue terminal',0
  db 'Quit terminal',0
;
;Set port menu
;
porttitle db '  Select port to use (now ',0, ')',0
;
snum1 = $
  db 'COM1',0
snum2 = $
  db 'COM2',0
snum3 = $
  db 'COM3',0
snum4 = $
  db 'COM4',0
snum5 = $
  db 'COM5',0
snum6 = $
  db 'COM6',0
snum7 = $
  db 'COM7',0
snum8 = $
  db 'COM8',0
snum9 = $
  db 'Leave alone',0
;
portmn label word
  dw snum1
  dw snum2
  dw snum3
  dw snum4
  dw snum5
  dw snum6
  dw snum7
  dw snum8
  dw snum9
  dw NULL
;
;Baud rate menu
;
baudtitle db '  Select Baud rate (now ',0, ')',0
;
snum1 = $
  db '110 Baud',0
snum2 = $
  db '150 Baud',0
snum3 = $
  db '300 Baud',0
snum4 = $
  db '600 Baud',0
snum5 = $
  db '1200 Baud',0
snum6 = $
  db '2400 Baud',0
snum7 = $
  db '4800 Baud',0
snum8 = $
  db '9600 Baud',0
snum9 = $
  db 'Leave alone',0
;
baudmn label word
  dw snum1
  dw snum2
  dw snum3
  dw snum4
  dw snum5
  dw snum6
  dw snum7
  dw snum8
  dw snum9
  dw NULL
;
;Parity menu
;
partitle db ' Select Parity (now ',0, ')',0
;
snum1 = $
  db 'None',0
snum2 = $
  db 'Odd',0
snum3 = $
  db 'Even',0
snum4 = $
  db 'Mark',0
snum5 = $
  db 'Space',0
snum6 = $
  db 'Leave alone',0
;
parmn label word
  dw snum1
  dw snum2
  dw snum3
  dw snum4
  dw snum5
  dw snum6
  dw NULL
;
;Data bits menu
;
dabtitle   db '  Select Word Length, is Number Data Bits (now ',0, ')',0
;
snum1 = $
  db '5 Bits',0
snum2 = $
  db '6 Bits',0
snum3 = $
  db '7 Bits',0
snum4 = $
  db '8 Bits',0
snum5 = $
  db 'Leave alone',0
;
dabmn label word
  dw snum1
  dw snum2
  dw snum3
  dw snum4
  dw snum5
  dw NULL
;
;Stop Bits menu
;
stotitle  db '  Select Number Stop Bits (now ',0, ')',0
;
snum1 = $
  db '1 Bit',0
snum2 = $
  db '2 Bits',0
snum3 = $
  db 'Leave alone',0
;
stomn label word
  dw snum1
  dw snum2
  dw snum3
  dw NULL
;
; End of data segment
  data_size equ $			;Size data segment
data ends
;
SUBTTL Main Program Code
PAGE
;
;Define code segment
code segment para public 'code'
;
;
;************************************************************************
;*									*
;*   Aterm main program. Sets up terminal (possibly using an external	*
;*	file of commands). Alternately polls keyboard and RS-232 port	*
;*	and sends out and displays respectively any charcters read.	*
;*									*
;************************************************************************
;
start proc far
;
;Initialise segment registers.
  mov ax,data
  mov ds,ax			;Initialise data segment register
assume ds:data
  mov psp,es			;Save location of PSP-used elsewhere
  mov es,ax			;Initialise extra segment register
;
assume es:data, ss:stack, cs:code
;
  call exec_defaults		;Use terminal definition file

  call open_com 		;Open serial port

mainloop:			;Main program loop

  call sendout			;Get characters/commands from keyboard

  call readin			;Get characters from port

  jmp short mainloop		;End of main loop

start endp
;
SUBTTL Send and receive high-level functions.
PAGE
;
;************************************************************************
;*									*
;*  Sendout- polls keyboard for either a character to be output or	*
;*	     a command. The latter use extended codes, which begin	*
;*	     with a zero byte. This routine calls the command/menu	*
;*	     routine when it encounters a zero byte.			*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Segment registers maintained
;   Flags undefined
;   All other regs undefined
;   Stack useage-substantial
;
sendout proc near
;
;Check keyboard
  call kbdnowait
  jnc snd01				;Exit if no character read
  or al,al
  jnz snd02				;is a character to ouput
					;is a command
  call cmdkey				;process
  jmp short snd01			;done

;Output character
snd02:
  test echo,echos			;Check echo
  jz snd03				;No echo to screen
					;Echo to screen
  mov xiobuf,al 			;Store in buffer
  xor al,al
  mov xiobuf+1,al			;nul terminate
  mov ax, fromecho
  mov bx, offset xiobuf
  call screen				;Output
  mov al,xiobuf 			;Get character back

snd03:
  mov si,offset outxlt			;output
  call xlt				;translation

  or cx,cx
  jz snd01				;Skip if translates to no characters

  call outstr				;Send charaters out

snd01:
  ret					;done
sendout endp
;
;
;************************************************************************
;*									*
;*  Readin- checks for any characters received at serial port.		*
;*	    Translates these and sends them to screen.			*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   Other regs undefined
;
; On exit
;
;   Flags undefined
;   Segment registers maintained
;   All other regs undefined.
;   Stack useage-up to 14 bytes plus DOS services
;
;
readin proc near
  call inpstr			;Read characters
  jnc rdi01			;skip if none read

  mov al,xiobuf 		;Get character
  mov si,offset inxlt		;input
  call xlt			;translate

  mov al,0
  mov [di],al			;Nul-terminate characters

  mov ax,fromport
  mov bx,offset xiobuf
  call screen			;display character

rdi01:
  ret				;done
readin endp
;
;
SUBTTL Process command keys/command strings
PAGE
;
;************************************************************************
;*									*
;*   Cmdkey- called when an extended key code has been received.	*
;*	     This gets remainder of key code and acts upon it.		*
;*									*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   segment registers maintained
;   All other regs undefined
;   Stack useage-substantial
;
cmdkey proc near
  call kbdnowait			;Look for remainder of extended code

  jnc cmk01				;exit if no key

  mov di,offset cmdkeytable		;Table of keys for commands
  mov cx, numcmdkey
  cld					;process forward

  repne scasb				;Search for matching character

  jne cmk01				;Skip if not a command key

  sub di,offset cmdkeytable + 1 	;Generate index
  shl di,1				;convert to word index
  mov ax,cmdkeyproc[di] 		;Get address of routine
  call ax				;execute it

cmk01:
  ret
cmdkey endp
;
;************************************************************************
;*									*
;*  Cliout- reads characters from keyboard (stdin) up to newl.		*
;*	    Line is written to xiobuf, and cmdout is called to		*
;*	    execute the command.					*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   Other register undefined.
;
; On exit
;
;   Flags undefined
;   Segment registers maintained
;   All other regs undefined
;   Stack useage-substantial
;
;
cliout proc near
  mov ax,fromctrl
  mov bx,offset cliprompt
  call screen				;Prompt for command line

  mov al,xiolen
  mov xiohead,al			;Set length parameter
  mov dx, offset xiohead		;Buffer address
  mov ah,0AH				;DOS buffered keyboard input
  int dos

  mov ax,fromctrl
  call screenewl			;Newline at end of input

  mov al,xiohead + 1			;Actual number entered
  xor ah,ah
  mov di,ax
  mov al,nul
  mov xiobuf[di],al			;Terminate input line

  call cmdout				;Attempt to execute command
  ret
cliout endp
;
;
;************************************************************************
;*									*
;*  Cmdout- Decode and execute command string in xiobuf.		*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   Other registers undefined
;   xiobuf = nul-terminated command string
;
; On exit
;
;   Flags undefined
;   segment registers maintained
;   All other regs affected
;   Stack useage-substantial, possibly recursive commands executed
;
cmdout proc near

  mov si,offset xiobuf
  call stripspace
  jc cmo99

  mov al,[si]				;Get first letter
  inc si				;move pointer to next character

  or al,tolower 			;Convert to lower case.
  mov di,offset cmdtable		;Table of commands (single letters)
  mov cx,numcmd
  cld					;process forward

  repne scasb				;Search for matching character

  jne cmo01				;Error if not a command

  sub di,offset cmdtable + 1		;Generate index
  shl di,1				;convert to word index
  mov ax,cmdproc[di]			;Get address of routine
  call ax				;execute it
  jmp short cmo99			;done

cmo01:
  mov ax,fromctrl
  mov bx,offset badcmdmes		;Signal bad command
  call screen
  mov bx,offset xiobuf
  call screen				;Print command
  call screenewl

cmo99:					;exit cmdout
  ret
cmdout endp
;
;
SUBTTL Setup menus
PAGE
;************************************************************************
;*									*
;*  Setup-presents user with a series of menus to set up the serial	*
;*	  port. 							*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage- 18 bytes + DOS services
;
;
setup proc near
sup01:
  call	sumain				;Main menu

  or al,al				;Decode user selection
  jnz sup02
;
;Option 0: Set COM port
;
  mov si,offset porttitle		;Build title for set port menu
  mov di,offset mainspace		;use mainspace
  mov bx,di
  call str_cpy
  dec di				;Back up over nul
  push si
  mov si,offset comname
  call str_cpy				;Get current com port name
  dec di				;Back up over nul
  pop si
  call str_cpy
  mov si,offset portmn
  call sumenu				;do menu
  cmp al,8
  jae sup09				;8 is option leave alone
  xor ah,ah
  shl ax,1
  mov si,ax
  mov si,portmn[si]			;get that option
  call set_com				;use as operand for set_com function.

sup09:					;Done setting port
  jmp short sup01

sup02:
  dec al
  jnz sup03
;
;Option 1: Set baud rate
;
  mov si,offset baudtitle		;Build title for set Baud rate menu
  mov di,offset mainspace		;use mainspace
  mov bx,di
  call str_cpy
  dec di				;Back up over nul
  push si
  push di
  mov di,offset baudnames
  mov ah,baud
  call skip_tostring
  mov si,di
  pop di
  call str_cpy				;Get current Baud rate
  dec di				;Back up over nul
  pop si
  call str_cpy
  mov si,offset baudmn
  call sumenu				;do menu
  cmp al,8
  jae sup10				;8 is option leave alone
  xor ah,ah
  shl ax,1
  mov si,ax
  mov si,baudmn[si]			;get that option
  call set_baud 			;use as operand for set_baud function.
sup10:					;Done setting Baud rate
  jmp short sup01

sup03:
  dec al
  jnz sup04
;
;Option 2: Set parity
;
  mov si,offset partitle		;Build title for set parity menu
  mov di,offset mainspace		;use mainspace
  mov bx,di
  call str_cpy
  dec di				;Back up over nul
  push si
  push di
  mov di,offset parynames
  mov ah,parity
  call skip_tostring
  mov si,di
  pop di
  call str_cpy				;Get current parity
  dec di				;Back up over nul
  pop si
  call str_cpy
  mov si,offset parmn
  call sumenu				;do menu
  cmp al,5
  jae sup11				;5 is option leave alone
  xor ah,ah
  shl ax,1
  mov si,ax
  mov si,parmn[si]			;get that option
  call set_parity			;use as operand for set_parity function.
sup11:					;Done setting parity
  jmp sup01

sup04:
  dec al
  jnz sup05
;
;Option 3: Set data bits
;
  mov si,offset dabtitle		;Build title for set data bits menu
  mov di,offset mainspace		;use mainspace
  mov bx,di
  call str_cpy
  dec di				;Back up over nul
  mov al,databits
  add al,dabbasechar
  mov [di],al				;get current data bits
  inc di				;move on
  call str_cpy
  mov si,offset dabmn
  call sumenu				;do menu
  cmp al,4
  jae sup12				;4 is option leave alone
  xor ah,ah
  shl ax,1
  mov si,ax
  mov si,dabmn[si]			;get that option
  mov [si+1],byte ptr nul		;nul terminate after digit
  call set_databits			;use as operand for set_databits fn.
sup12:					;Done setting data bits
  jmp sup01

sup05:
  dec al
  jnz sup06
;Option 4: Set stop bits
;
  mov si,offset stotitle		;Build title for set stop bits menu
  mov di,offset mainspace		;use mainspace
  mov bx,di
  call str_cpy
  dec di				;Back up over nul
  mov al,stopbits
  add al,stobasechar
  mov [di],al				;get current stop bits
  inc di				;move on
  call str_cpy
  mov si,offset stomn
  call sumenu				;do menu
  cmp al,2
  jae sup13				;2 is option leave alone
  xor ah,ah
  shl ax,1
  mov si,ax
  mov si,stomn[si]			;get that option
  mov [si+1],byte ptr nul		;nul terminate after digit
  call set_stopbits			;use as operand for set_stopbits fn.
sup13:					;Done setting stop bits
  jmp sup01

sup06:
  dec al
  jnz sup07
;Option 5: Continue terminal session
  mov ax,fromctrl
  mov bx,offset supm1				;say goodbye, and
  call screen
  jmp short sup08				;done

sup07:
;Option 6: Quit terminal
  call quit					;quit

sup08:
  ret						;done
setup endp
;
;
;************************************************************************
;*									*
;*  Sumain- Generates text of main menu and then calls sumenu to get	*
;*	    option. Text includes corrent settings of parameters	*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-16 bytes + DOS services
;
;
sumain proc near
  mov di,offset mainspace		;Area to build main menu
  mov si,offset mainmn			;Source of text for menu
  mov bx,offset mainptspace		;area for pointers


  xchg bx,di
  mov [di],bx				;Store address port line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;1st half port option
  dec di				;backup over nul
  push si
  mov si,offset comname
  call str_cpy				;Port name
  dec di				;backup over nul
  pop si
  call str_cpy				;2nd half port option

  xchg bx,di
  mov [di],bx				;Store address Baud line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;1st half baud option
  dec di				;backup over nul
  push si
  push di
  mov di,offset baudnames
  mov ah,baud
  call skip_tostring
  mov si,di
  pop di
  call str_cpy				;Baud name
  dec di				;backup over nul
  pop si
  call str_cpy				;2nd half Baud option

  xchg bx,di
  mov [di],bx				;Store address parity line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;1st half parity option
  dec di				;backup over nul
  push si
  push di
  mov di,offset parynames
  mov ah,parity
  call skip_tostring
  mov si,di
  pop di
  call str_cpy				;Parity name
  dec di				;backup over nul
  pop si
  call str_cpy				;2nd half Parity option

  xchg bx,di
  mov [di],bx				;Store address databits line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;1st half databits option
  dec di				;backup over nul
  mov al,databits
  add al,dabbasechar			;Convert to correct character
  mov [di],al
  inc di				;move on
  call str_cpy				;2nd half databits option

  xchg bx,di
  mov [di],bx				;Store address stopbits line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;1st half stopbits option
  dec di				;backup over nul
  mov al,stopbits
  add al,stobasechar			;Convert to correct character
  mov [di],al
  inc di				;move on
  call str_cpy				;2nd half stopbits option

  xchg bx,di
  mov [di],bx				;Store address continue line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;continue option

  xchg bx,di
  mov [di],bx				;Store address quit line
  inc di
  inc di
  xchg bx,di

  call str_cpy				;quit option

  mov di,bx
  mov [di],word ptr NULL		;end of pointers

  mov si,offset mainptspace
  mov bx,offset mainname
  call sumenu				;Get option

  ret
sumain endp
;
;
;************************************************************************
;*									*
;*  Sumenu-menu driver for setup. Displays menu specified and gets	*
;*	   user choice							*
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = Pointer to menu description table
;   BX = pointer to title of menu
;   Segment registers as start proc
;
; On exit
;
;   AL = option selected (counts from 0 for first option)
;   Flags undefined
;   All other regs maintained.
;   Stack useage-14 bytes + DOS services
;
;
sumenu proc near
  mov ax,fromctrl
  call screenewl
  call screenewl
  call screen					;print title

  mov bx,offset smum1				;Header for table
  call screen

  mov cx,1					;option counter
smu01:
  mov bx,[si]
  cmp bx,NULL
  je smu03					;Exit on pointer = NULL
  push bx
  mov bx,offset smum2
  call screen
  mov bx,cx
  call wrint					;option number
  mov bx,offset smum3
  call screen
  pop bx					;access menu option name
  call screen
  inc si
  inc si
  inc cx
  jmp short smu01				;next option

smu03:
  mov bx,offset smum4
  call screen

  dec cx					;Normalise
  dec cx
smu02:
  call kbdwait					;Get user option
  sub al,'1'
  cmp al,cl
  ja smu02					;Try again if option bad

  push ax
  add al,'1'
  mov bl,al
  mov ax,fromctrl
  call screenchar				;Print option
  call screenewl
  pop ax
  ret
sumenu endp
;
;
SUBTTL Find and execute defaults file
PAGE
;************************************************************************
;*									*
;*  Exec_defaults-look for command line parameter and execute as	*
;*		  definition file. If not present or a bad file 	*
;*		  then try to execute def_file. 			*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-substantial
;
;
exec_defaults proc near
  mov ds,psp				;Access PSP via DS
  assume ds:nothing
  mov si,080H				;source bytes
  lodsb
  mov cl,al				;Get number bytes in parameter(s) on
  xor ch,ch				;command line
  push cx				;save number
  mov di,offset xiobuf			;destination
  rep movsb				;Get parameter(s)
  mov ax,es
  mov ds,ax				;resore data segment
  assume ds:data
  xor al,al
  stosb 				;nul-terminate
  pop cx				;get back number
  or cx,cx
  jz exdf01				;no parameters

  mov si,offset xiobuf
  call exec_file			;Try to exec parameter as filename
  jnc exdf02				;Done if that succedes
exdf01:
  mov si,offset def_file
  call exec_file			;Attempt to execute default file
  jnc exdf02				;Done if that succedes
;No file of settings run. Need to set port to default values
;via BIOS:
  call bios_x_init

exdf02:
  ret
exec_defaults endp
;
;
SUBTTL Command execution functions.
PAGE
;************************************************************************
;*									*
;*  Exec_dos- uses parameter string as command to send to command.com	*
;*									*
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameters.
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 bytes + DOS services
;
;
exec_dos proc near
  call stripspace
  jc exd02				;empty command line ==> let command.com
					;prompt user.
;format parameter area for use by command.com
  dec si				;note that 3 header bytes before xiobuf
					;allow us to do this
  mov [si],byte ptr spce		;At least one space before parameters
  dec si
  dec si
  mov [si],word ptr 'C/'                ;command line switch '/C'
  mov di,si
  dec si
  xor al,al
  mov cx,-1
  cld
  repne scasb				;look for nul termination of string
  dec di
  mov [di],byte ptr cr			;change to cr termination
  sub di,si				;find length of command line
  mov ax,di
  mov [si],al				;store length
  jmp short exd03

exd02:		;set empty command line - no /c switch
  mov [si],byte ptr cr
  dec si
  mov [si],byte ptr 0			;no bytes in string
  mov ax,fromctrl
  mov bx,offset exdm1
  call screen				;exit message

exd03:
  mov exdcmdline,si			;offset command line in parameter block

  mov ax,psp
  mov exdpsp1,ax			;set up other parts of parameter block
  mov exdpsp2,ax

;set memory allocation- first look for highest segment in memory
  mov bx,stack
  mov ax,offset stack_size
  cmp bx,data
  ja exd05
  mov bx,data
  mov ax,offset data_size
exd05:
  cmp bx,code
  ja exd06
  mov bx,code
  mov ax,offset code_size
exd06:				;found highest segment, use to generate total
				;length of program
  mov cl,4
  shr ax,cl				;Number paragraphs
  add bx,ax
  sub bx,psp				;start of program
  inc bx				;Be on safe side

  mov ax,psp
  mov es,ax				;segment address memory
  assume es:nothing
  mov ah,04AH				;DOS Set Block
  int dos
  mov ax,ds
  mov es,ax				;restore ES
  assume es:data
  jnc exd04
					;Memory reallocation failed
  mov ax,fromctrl
  mov bx,offset exdfm2
  call screen
  jmp short exd01			;exit

exd04:
  mov ax,psp
  mov es,ax
  assume es:nothing			;Access PSP
  mov di,02CH
  mov ax,es:[di]			;get address of environment
  mov es,ax				;access environment
  mov di,0				;offset into environment
  mov si,offset comspecname		;'COMSPEC='
  call find_env 			;find string named 'COMSPEC='
  jnc exd07				;(is path of command.com)

  mov ax,ds
  mov es,ax
  assume es:data			;restore ES
  mov ax,fromctrl
  mov bx,offset exdfm3
  call screen
  jmp short exd01

exd07:

  mov bx,offset exdparams		;address parameteter block
  mov dx,di				;Command.com path name
  mov ax,es
  mov ds,ax				;segment of path name
  assume ds:nothing
  mov ax,data
  mov es,ax				;segment of parameter block
  assume es:data
  mov ax,04B00H 			;DOS load and execute program
  int dos
  mov ax,es
  mov ds,ax
  assume ds:data			;Restore DS
  jnc exd01
  mov bx,ax
  mov ax,fromctrl
  mov bx,offset exdfm1			;Failed to run COMMAND.COM
  call screen

exd01:					;done
  ret
exec_dos endp
;
;
;************************************************************************
;*									*
;*  Exec_file-Scans parameters for a path name, attempts to open	*
;*	      file and exucte the text in it as a series of commands	*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameters
;   Segment registers as start proc
;
; On exit
;
;   carry = true  ==> could not find file/other failure
;   carry = false ==> found file
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-substantial, file may execute other files
;
;
exec_file proc near
  call stripspace
  jc exf01					;No name
  mov dx,si					;Save start path name
  call skiptospace
  call stripspace
  jnc exf02					;To many names

  mov ax,03D00H 				;Access code 0 ==> read
						;DOS service 3DH, open handle
  int dos
  jc exf04					;failed to open file

  push ax					;save handle

  mov ax,fromctrl				;print message
  mov bx,offset exfm1
  call screen
  mov bx,dx					;file path
  call screen
  call screenewl

;read in line from file
  pop bx					;handle
exf06:
  mov cx,xiolen 				;Max length of buffer
  call readline
  cmp cx,-1					;eof marker
  je exf08					;done executing file.
  or cx,cx
  jz exf06					;empty line

  push bx					;save handle
  call cmdout					;execute command
  pop bx					;recover handle
  jmp short exf06				;get next line

exf04:						;Could not find file
  mov ax,fromctrl
  mov bx,offset exffm3
  call screen
  mov bx,dx					;Print path name
  call screen
  call screenewl
  jmp short exf03

exf02:						;Two many file names
  mov ax,fromctrl
  mov bx,offset exffm2
  call screen
  jmp short exf03

exf01:						;No file name
  mov ax,fromctrl
  mov bx,offset exffm1
  call screen

exf03:
  stc						;Failed
  jmp short exf07

exf08:
  clc						;succeded
exf07:
  ret						;Done
exec_file endp
;
;
;************************************************************************
;*									*
;*  Readline-Attempt to read line of text from handle in BX. Reads	*
;*	     up to first of cr or lf, and nul-terminates string.	*
;*	     Line is written to xiobuf, up to xiolen characters inc	*
;*	     terminating nul. CX returns number characters in line	*
;*	     or -1 for end of file.					*
;*									*
;************************************************************************
; On entry
;
;   BX = File handle to use
;   Segment registers as start proc
;
; On exit
;
;   CX = -1   ==> end of file reached or
;   CX = number of characters read into buffer (excluding terminating nul)
;   Flags undefined
;   AX,DX,SI undefined
;   All other regs maintained
;   Stack useage-as DOS service 03FH
;
;
readline proc near
  mov dx,offset xiobuf				;pointer into buffer
  xor si,si					;counter for bytes read
  mov cx,1					;read one byte at a time
rdl01:
  mov ah,03FH					;DOS read handle
  int dos
  jc rdl02					;Error-return eof
  or ax,ax
  je rdl02					;EOF
  mov al,xiobuf[si]				;get byte
  cmp al,26
  je rdl02					;end of file CTRL-Z
  cmp al,cr
  je rdl04					;end of line
  cmp al,lf
  je rdl04					;end of line
  cmp si,xiolen - 1				;Don't update if at end
  jae rdl05					;of buffer
  inc dx
  inc si
rdl05:
  jmp short rdl01

rdl02:						;error/end of file
  or si,si
  jnz rdl04					;return any characters read
						;Next readln will return eof
  mov si,-1					;None read and eof

rdl04:
  mov cx,si					;number of byte read in
  xor al,al
  mov xiobuf[si],al				;nul-terminate line

  ret
readline endp
;
;
;************************************************************************
;*									*
;*  Transmit-Scans operand to build a sequence of characters. These	*
;*	     are sent (without translation) to serial port.		*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to  operands
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-12 bytes + DOS services
;
;
transmit proc near
  push si
  mov di,offset setxbuf 		;Destination for parse
  call parse_str_def
  jc tns01				;bad parameter
  lodsb
  or al,al
  jnz tns02				;More than one parameter

  mov ax,fromctrl			;print sending message
  mov bx,offset tnsm1
  call screen
  pop bx				;Address parameter sequence
  call screen
  call screenewl

  push cx
  mov si,offset setxbuf
  mov di,offset xiobuf
  rep movsb				;Copy sequence to o/p buffer
  pop cx
  call outstr				;send characters

  jmp short tns03

tns01:					;Bad parameter
  pop ax				;clear stack
  mov ax,fromctrl
  mov bx,offset tnsfm1
  call screen
  jmp short tns03			;exit

tns02:					;More than one parameter
  pop ax				;clear stack
  mov ax,fromctrl
  mov bx,offset tnsfm2
  call screen

tns03:
  ret					;done
transmit endp
;
;
;************************************************************************
;*									*
;*  Send_comment-Sends text of comment to screen.			*
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to comment text
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   BX,AX undefined
;   All other regs maintained.
;   Stack useage-12 bytes + DOS services
;
;
send_comment proc near
  mov bx,si
  mov ax,fromctrl
  call screen
  call screenewl
  ret
send_comment endp
;
;
;************************************************************************
;*									*
;*  Set_com-reads parameters for the name of a device to use as serial	*
;*	    port. If is of form 'COMn' then scan in n as a number.      *
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameters
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-12 bytes + DOS services
;
;
set_com proc near
  call str_toupper			;Set to upper case
  call stripspace
  jnc sec06
  jmp sec01
sec06:
  mov dx,si
  call skiptospace			;find end string
  mov cx,si
  mov si,dx				;recover pointer
  sub cx,si				;length of string
  cmp cx,8
  jbe sec04
  mov ax,fromctrl
  mov bx,offset secwm1			;Warn extra characters skipped
  call screen
sec04:
  mov di,offset comname
  rep movsb				;copy over name
  xor al,al
  stosb 				;nul-terminate

  call stripspace			;look for other parameters
  jc sec02
  mov ax,fromctrl
  mov bx,offset secwm2
  call screen				;Extra parameters skipped

sec02:
;Look for a name of form 'COMn', and get number n
  cmp comname,byte ptr 'C'
  jne sec03				;Not 'Cxx'
  cmp comname + 1,byte ptr 'O'
  jne sec03				;Not 'COx'
  cmp comname + 2,byte ptr 'M'
  jne sec03				;Not 'COM'

;Name is of form 'COM..'
  mov si,offset comname + 3		;start of number
  call scanint				;check it out
  jc sec03				;Not COMn
  dec ax
  cmp ax,15
  ja sec03				;n > 15
  mov bx,si
  call skiptospace
  cmp si,bx
  jne sec03				;Number not last part of name

  mov comno,ax				;is good com port name
					;at last!
  jmp short sec07

sec03:
  mov comno,-1				;Not a good port number
sec07:
  mov ax,fromctrl
  mov bx,offset secm1
  call screen				;Say what happening
  mov bx,offset comname
  call screen
  call screenewl

  call close_com			;Close com port
  call open_com 			;open com port with new name
  jmp short sec05			;done

sec01:
  mov ax,fromctrl
  mov bx,offset secfm1
  call screen				;Bad name string

sec05:
  ret
set_com endp
;
;
;************************************************************************
;*									*
;*  Set_echo-read parameters, set echo variable in accordance		*
;*	     parameters are any number of single letters 'n' for no     *
;*	     echo, 's' for screen echo and 'l' for line echo.           *
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameters.
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   AX,BX,SI,CX undefined
;   All other regs maintained
;   Stack useage-12 bytes + DOS services
;
;
set_echo proc near
see03:
  mov echo,echon
see01:
  call stripspace
  jc see02				;End of line ==> done
  mov al,[si]				;get character
  inc si
  or al,tolower
  cmp al,'n'
  je see03				;Set no echo

  cmp al,'l'
  jne see04
  or echo,echol 			;Set echo to line
  jmp short see01

see04:
  cmp al,'s'
  jne see07
  or echo,echos 			;set echo to screen
  jmp short see01

see07:
  mov ax,fromctrl
  mov bx,offset seefm1
  call screen				;Bad character message
  jmp short see01

see02:					;Done reading parameters.
					;print results
  mov ax,fromctrl
  mov bx,offset seem1
  call screen

  mov bx,echo				;skip to correct name of echo
  mov di,offset echonames
  call skip_tostring
  mov bx,di
  mov ax,fromctrl
  call screen				;print name
  call screenewl

  ret
set_echo endp
;
;
;************************************************************************
;*									*
;*  Set_stopbits-Decode parameter and use to set number stop bits.	*
;*		 Parameter is a number, either 1 or 2			*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameter string
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-14 bytes + DOS services
;
;
set_stopbits proc near
  call stripspace
  jc sst01
  call scanint					;Get number
  jc sst01
  dec ax					;convert to parameter
  cmp ax,1					;Check number
  ja sst01
  call stripspace
  jnc sst01					;Other charatcers on line

;Number ok
  mov stopbits,al
  call bios_x_init				;Attempt to set
  jc sst03					;failed to set

  mov ax,fromctrl
  mov bx,offset sstm1
  call screen
  mov bl,stopbits
  inc bl					;Make up to equal num bits.
  xor bh,bh
  call wrint					;print number.
  call screenewl
  jmp short sst02

sst01:
  mov ax,fromctrl
  mov bx,offset sstmf1
  call screen				;Bad specification
  jmp short sst02

sst03:
  mov ax,fromctrl
  mov bx,offset sstmf2
  call screen				;failed to set

sst02:
  ret					;done
set_stopbits endp
;
;************************************************************************
;*									*
;*  Set_databits-Decode parameter and use to set number data bits.	*
;*		 Parameter is a number in range 5..8			*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameter string
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-14 bytes + DOS services
;
;
set_databits proc near
  call stripspace
  jc sdb01
  call scanint					;Get number
  jc sdb01
  sub al,5					;convert to parameter
  cmp ax,3					;Check number
  ja sdb01
  call stripspace
  jnc sdb01					;Other charatcers on line

;Number ok
  mov databits,al
  call bios_x_init				;Attempt to set
  jc sdb03					;failed to set

  mov ax,fromctrl
  mov bx,offset sdbm1
  call screen
  mov bl,databits
  add bl,5					;Make up to equal num bits.
  xor bh,bh
  call wrint					;print number.
  call screenewl
  jmp short sdb02

sdb01:
  mov ax,fromctrl
  mov bx,offset sdbmf1
  call screen				;Bad specification
  jmp short sdb02

sdb03:
  mov ax,fromctrl
  mov bx,offset sdbmf2
  call screen				;failed to set

sdb02:
  ret					;done
set_databits endp
;
;
;************************************************************************
;*									*
;*  Set_parity-Decode operand string for a parity specifier.		*
;*	       These are: 'n', 'o', 'e', 'm' and 's'. Case is ignored   *
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to parameter string
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 bytes + DOS services
;
;
;
set_parity proc near
  call stripspace
  jc spy01				;Nul encountered

  mov al,[si]				;get parameter
  or al,tolower 			;Set to lower case

  mov di,offset paritychars
  mov cx,numparych
  repne scasb				;look for matching parity
  jne spy01
  mov parity,cl 			;store parity
  call bios_x_init			;set it
  jc spy03

  mov ax,fromctrl
  mov bx,offset spym1
  call screen

  mov ah,parity 			;skip to correct name of parity
  mov di,offset parynames
  call skip_tostring
  mov bx,di
  mov ax,fromctrl
  call screen
  mov bx,offset spym2
  call screen
  jmp short spy02

spy03:
  mov ax,fromctrl
  mov bx,offset spyf2
  call screen				;Cannot set parity
  jmp short spy02

spy01:
  mov ax,fromctrl
  mov bx,offset spyf1
  call screen				;Bad parameter message
  jmp short spy02

spy02:
  ret
set_parity endp
;
;
;************************************************************************
;*									*
;*  Quit- Executes DOS service 4CH, end process, with return code 0.	*
;*									*
;************************************************************************
; On entry
;
;   registers undefined
;
; On exit
;
;   (inapplicable)
;
quit proc near

  mov ax,fromctrl
  mov bx,offset quitmes
  call screen				;Say quitting.

  mov ax,4C00H				;DOS end process, return 00
  int dos
  ret
quit endp
;
;
;************************************************************************
;*									*
;*  Help- Display help information on screen.				*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   AX undefined
;   BX undefined
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 bytes + DOS services
;
help proc near
  mov ax,fromctrl
  mov bx,offset helpmes
  call screen				;Print help message
  ret
help endp
;
;
;************************************************************************
;*									*
;*  set_baud-read number and use to set baud rate. Ouputs error if	*
;*	     not a valid baud rate. Baud rates > 65k are not checked	*
;*	     exactly.							*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to operands of baud command
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   Segment regs maintained.
;   Other regs undefined
;   Stack useage-12 bytes + service
;
set_baud proc near
  call stripspace
  jc sbd01				;no parameters at all
  call scanint				;look for baud rate
  jc sbd01				;Not a number

  mov bh,baud
  push bx				;save baud rate

  call prefer_baud			;Convert baud rate to BIOS parameter
  mov baud,al
  call bios_x_init			;attempt set parameters
  jc sbd02
  pop ax				;good baud rate

;print out message
  mov ax,fromctrl
  mov bx,offset setbaud1
  call screen

  mov ah,baud				;skip to correct name of baud rate
  mov di,offset baudnames
  call skip_tostring
  mov bx,di
  mov ax,fromctrl
  call screen
  call screenewl
  jmp short sbd04			;done

sbd02:
  pop ax				;clean stack
  mov baud,ah
sbd01:
  mov ax,fromctrl
  mov bx,offset badbaud
  call screen
sbd04:
  ret					;done
set_baud endp
;
;************************************************************************
;*									*
;*  Set_inxlt-Set the input translation specified.			*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to operands
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   AX,BX,CX,SI,DI undefined
;   All other regs maintained.
;   Stack useage-16 bytes + DOS services
;
;
set_inxlt proc near
  mov ax,fromctrl
  mov bx,offset sixm1
  call screen					;message

  mov di,offset inxlt				;set correct variables
  call set_xlt
  ret
set_inxlt endp
;
;
;************************************************************************
;*									*
;*  Set_outxlt-Set the output translation specified.			*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to operands
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   AX,BX,CX,SI,DI undefined
;   All other regs maintained.
;   Stack useage-16 bytes + DOS services
;
;
set_outxlt proc near
  mov ax,fromctrl
  mov bx,offset soxm1
  call screen					;message

  mov di,offset outxlt				;set correct variables
  call set_xlt
  ret
set_outxlt endp
;
;
;************************************************************************
;*									*
;*  set_xlt-Set a translation. Crys foul if translation specification	*
;*	   is bad. Format of specification is				*
;*		<str1>:<strn>						*
;*	   where							*
;*		<str1> = <strn> of one character long			*
;*		<strn> = {|<char>|<quoted_string>}[,<strn>]		*
;*		<char> = decimal number 				*
;*		<quoted_string> = '0 or more characters'                *
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to operands of set_xlt command
;   DI = pointer to translation table to update
;   Segment registers as start proc
;
;
; On exit
;
;   Flags undefined
;   AX,BX,CX,SI,DI undefined
;   All other regs maintained.
;   Stack useage-14 bytes + DOS services
;
;
;
set_xlt proc near
  push di					;Save translation table
  mov di,offset setxbuf 			;Store string
  call stripspace
  mov sxtkeyst,si				;save location of key

  call parse_str_def				;Look for key to translation
  jnc sxt01

  mov ax,fromctrl				;Bad string
  mov bx,offset sxtfm1
  call screen					;return
  jmp short sxt07

sxt01:
  cmp cx,1
  je sxt03
						;Key not a single character
  mov ax,fromctrl
  mov bx,offset sxtfm2
  call screen
  jmp short sxt07				;return

sxt03:						;good key string
  mov al,setxbuf
  push ax					;save key string
  cmp [si],byte ptr colon			;Check separator
  je sxt04

  mov ax,fromctrl				;Bad separator
  mov bx,offset sxtfm3
  call screen
  jmp short sxt02				;return

sxt04:
  xor al,al
  mov [si],al					;nul terminate
  inc si
  call stripspace
  mov sxttrans,si				;save location translation
  mov di,offset setxbuf
  call parse_str_def				;Look for translation
  jnc sxt05

  mov ax,fromctrl				;Bad translation
  mov bx,offset sxtfm4
  call screen
  jmp short sxt02				;return

sxt05:
  call stripspace
  jc sxt06					;Should be end of line

  mov ax,fromctrl				;Other non-space characters
  mov bx,offset sxtfm5
  call screen

sxt06:
  pop ax					;key character
  pop si					;translation table
  call setxltlow				;change translation.

  mov ax,fromctrl
  mov bx,offset sxtm1
  call screen
  mov bx,sxtkeyst
  call screen
  mov bx,offset sxtm2
  call screen
  mov bx,sxttrans
  call screen
  call screenewl
  jmp short sxt08				;done

sxt02:
  pop ax					;clear stack
sxt07:
  pop ax					;clear stack
sxt08:
  ret
set_xlt endp
;
;
;************************************************************************
;*									*
;*  Setxltlow-Sets a translation. Updates specified translation 	*
;*	      table. If there is not enough space left to accomodate	*
;*	      the new translation then output an error message and	*
;*									*
;************************************************************************
; On entry
;
;   AL = character key for translation
;   setxbuf = characters to be substituted for AL
;   CX = number characters in translation.
;   SI = translation table to use
;   Segment registers as start proc
;
; On exit
;
;   AX,BX,CX,DX,SI,DI undefined
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 + DOS services
;
setxltlow proc near
  push cx			;number characters in translation
  mov sxlxlt,si 		;Translation table

  mov dx,[si+(256*2)]		;Get pointer to end translation table

  sub dx,[si]			;subtract pointer to begining of table
				;-gives total length of table
  xor ah,ah
  inc ax
  shl ax,1
  add si,ax			;Generate pointer to end translation
  mov bx,[si]			;Get it
  mov sxleox,si 		;save end of translation
  dec si
  dec si			;Pointer to start of translation
  sub bx,[si]			;Generate length of old translation

  sub cx,bx			;generate change in length
  mov sxldlen,cx

  add dx,cx			;Get new total length of table
  cmp dx,xlt_size		;check it
  jbe sxl01

  pop ax			;loose stack data
  mov ax,fromctrl		;Error message not enough space
  mov bx,offset sxlfail
  call screen
  jmp short sxl02		;quit routine

sxl01:
  mov di,sxlxlt
  mov ax,[di+(256*2)]		;end old table
  mov di,sxleox 		;end old translation
  sub ax,[di]			;generate length block to move

;move representation up or down to make space for new substitution
  cmp sxldlen,0
  jle sxl03			;need to shuffle characters back
				;else need to shuffle them up to create space

  std				;direction is backwards
  mov si,sxlxlt
  mov si,[si+(256*2)]		;old end table
  dec si			;point to last character

  jmp short sxl04
sxl03:				;shuffle characters back to fil gap
  cld				;direction is forwards
  mov si,sxleox 		;end old tranlation
  mov si,[si]

sxl04:				;Do actual move
  mov di,si
  add di,sxldlen		;Generate destination via change in length
  mov cx,ax			;number bytes to move
  rep movsb

;Update pointers.
  mov si,sxlxlt
  add si,256*2
  mov ax,sxldlen

sxl05:
  add [si],ax			;move pointer
  dec si
  dec si
  cmp si,sxleox
  jae sxl05			;loop

;write in new representation
  mov si,sxleox
  dec si
  dec si			;pointer to pointer to start translation
  mov di,[si]			;dereference
  pop cx			;number characters in translation
  mov si,offset setxbuf
  cld
  rep movsb			;move bytes

sxl02:				;done
  ret
setxltlow endp
;
;
;************************************************************************
;*									*
;*  prefer_baud- converts word in AX to byte in AL, former is a 	*
;*		 (possibly truncated) baud rate, the latter is a	*
;*		 number suitable for use as a parameter in the BIOS	*
;*		 extended initialise.					*
;*									*
;************************************************************************
; On entry
;
;   AX = Baud rate
;   Segment registers as start proc
;
; On exit
;
;   AL = BIOS parameter, or -1 if baud rate not recognised.
;   CL,SI undefined
;   Flags undefined
;   All other regs maintained.
;   Stack useage-none
;
;
prefer_baud proc near
  xor cl,cl
  mov si,offset pfbtable		;Table of baud rates
pfb01:
  cmp [si],word ptr 0
  je pfb03				;end of sublist for this baud rate
  cmp ax,[si]
  je pfb02				;Done-found matching baud rate
  inc si
  inc si				;Next synonymn
  jmp pfb01

pfb03:
  inc cx				;Next Baud rate
  inc si
  inc si
  cmp [si],word ptr 0			;End of list
  jne pfb01				;No-next baud rate
  mov al,-1				;Bad baud rate

pfb02:
  mov al,cl				;Return Baud rate param.
  ret
prefer_baud endp
;
;
;************************************************************************
;*									*
;*  bios_x_init-check the global parameters for baud rate etc. If	*
;*		they are ok then use to set the current serial port.	*
;*		return carry set if fails.				*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   Globals comno,baud,parity,stopbits,startbits set with values
;
; On exit
;
;   Carry = true ==> values bad, could not set port
;     AH,BX,CX,DX undefined
;   Carry = false ==> values good, used to set port.
;     AX,BX,CX,DX as BIOS service 04H
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-as BIOS service (currently 22 bytes)
;
; Note- this service assumes the presence of Brain Boxes Asynchronous
;   BIOS. This is included in the Brain Boxes serial device driver.
;
;
bios_x_init proc near
  mov al,brk
  cmp al,1
  ja bxi01				;Bad break parameter
  mov bh,parity
  cmp bh,4
  ja bxi01				;Bad parity
  mov bl,stopbits
  cmp bl,1
  ja bxi01				;Bad stop bits
  mov ch,databits
  cmp ch,3
  ja bxi01				;Bad data bits
  mov cl,baud
  cmp cl,0BH
  ja bxi01				;Bad baud rate
  mov dx,comno
  cmp dx,15
  ja bxi01				;Bad port

  mov ah,04H				;BIOS extended initialise
  int com				;BIOS asynchronous services

  clc
  jmp short bxi02			;exit good

bxi01:
  stc					;exit bad
bxi02:
  ret
bios_x_init endp			;done
;
;
;
SUBTTL Serial port read/write/translate routines.
PAGE
;
;************************************************************************
;*									*
;*  xlt-routine translates the character in AL via table at SI		*
;*	translation written to xiobuf					*
;*									*
;************************************************************************
;
; On entry
;   AL = Character to translate.
;   SI = Address of table to use for translation.
;   Segment registers as start proc
;
; On exit
;   CX = number of characters in translation
;
;   AX = number of characters in translation
;   SI = pointer to character after end translation in tables
;   DI = pointer to character after end buffer.
;   xiobuf = translation
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-none
;
;
xlt proc near
  xor ah,ah
  inc ax
  shl ax,1
  add si,ax			;Generate pointer to end translation
  mov ax,[si]			;Get it
  dec si
  dec si			;Pointer to start of translation
  sub ax,[si]			;Generate length
  mov cx,ax
  mov si,[si]			;Set source
  mov di,offset xiobuf		;Set desination
  cld				;Process forward

  rep movsb			;Move characters across

  mov cx,ax			;restore count
  ret
xlt endp
;
;
;************************************************************************
;*									*
;*  Inpstr- attempts to read a character from the serial port. If	*
;*	    echo to line is set then echo character back to port.	*
;*	    Character is written to xiobuf.				*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;
; On exit
;
;   Carry = true ==> character has been read
;   Carry = false ==> no character read
;
;   AX undefined
;   BX undefined
;   CX undefined
;   DX undefined
;   xiobuf = character read
;
;   Other lags undefined
;   All other regs maintained.
;   Stack useage-12 bytes + DOS services
;
;
inpstr proc near
  mov ah,3FH				;DOS read file handle
  mov bx,comport			;File handle
  mov cx,1				;Read 1 character
  mov dx,offset xiobuf			;Where to store it
  int dos

  cmp ax,1				;Number read
  jne inst01				;None read

  mov ax,echo				;Check echo
  and ax,echol
  jz inst03				;No echo
					;echo
  mov cx,1				;1 character
  call outstr				;send it
inst03:
  stc					;Some read
  jmp short inst02

inst01:
  clc					;None read
inst02:
  ret					;done
inpstr endp
;
;
;************************************************************************
;*									*
;*  Outstr- sends characters from xiobuf to serial port. Reports an	*
;*	    error to user if this fails. The routine returns anyway.	*
;*									*
;************************************************************************
; On entry
;   CX = number bytes to send
;   Segment registers as start proc
;   xiobuf = characters to be sent
;
; On exit
;   AX undefined
;   BX undefined
;   CX undefined
;   DX undefined
;   xiobuf maintained
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 bytes + DOS services
;
outstr proc near
  mov ah,40H				;DOS write handle
  mov bx, comport			;File handle
  mov dx,offset xiobuf			;Where data
  int dos
  jnc outs01				;Data sent ok

  mov ax,fromctrl			;Flag source of output
  mov bx,offset writefail
  call screen				;Display error message

outs01:
  ret
outstr endp
;
;
SUBTTL Open/Close serial port routines
PAGE
;
;************************************************************************
;*									*
;*  Close_com-attempts to close file comport. Sets file handles to -1	*
;*									*
;************************************************************************
;
; On entry
;   Segment registers as start proc
;   Other regs undefined
;   variable comport = handle associated with serial port
;
; On exit
;   AX undefined
;   BX undefined
;
;   Flags undefined
;   All other regs maintained.
;   comport = -1
;   Stack useage-as DOS functions 3EH
;
;
close_com proc near
  mov bx,comport
  cmp bx,-1				;-1 means not set
  je clsp01

  mov ah,3EH				;DOS close handle
  int dos
  mov comport,-1			;Set unset status

clsp01:
  ret					;Done
close_com endp
;
;
;************************************************************************
;*									*
;*  open_com- attempts to open file comname. Stores file handle or -1	*
;*	      in comport. If fails prints a message to that effect.	*
;*									*
;************************************************************************
;
; On entry
;   Segment registers as start proc
;   Other regs undefined
;
; On exit
;   AX undefined
;   BX undefined
;   DX undefined
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-10 + DOS services
;
;
open_com proc near
  mov ah,3DH			;DOS open handle
  mov al,2			;Access code ==> no inherit, compatibility
				;sharing and read/write access.
  mov dx, offset comname	;Name of device
  int dos
  jnc opnc01			;Carry clear means no error.
;deal with error
  mov ax, fromctrl		;Print out error message
  mov bx, offset openfail1
  call screen
  mov bx, offset comname
  call screen
  mov bx, offset openfail2
  call screen

  mov ax,0FFFFH 		;Flag bad port
  mov comport,ax
  jmp short opnc02		;exit

opnc01:
  mov comport,ax		;Save file handle
  call uncook			;Set file to raw data mode
opnc02:
  ret				;Done
open_com endp
;
;
;************************************************************************
;*									*
;*  uncook-routine sets mode of file to raw data. I/O is performed	*
;*	   as many bytes as possible at a time, and control chars	*
;*	   are not trapped.						*
;*									*
;************************************************************************
;
; On entry
;   AX = valid file handle (note this routine does no checking)
;
; On exit
;   AX undefined
;   DX undefined
;   BX = file handle
;
;   Flags undefined
;   All other regs maintained.
;   Stack useage-as DOS functions 44H
;
;
uncook proc near
  mov bx,ax
  mov ah,44H			;DOS IOCTL
  mov al,0			;Get device data
  int dos

  xor dh,dh
  or dl,00100000b		;Set bit 5 of data
  mov ah,44H			;DOS IOCTL
  mov al,1			;Set device data
  int dos

  ret
uncook endp
;
;
SUBTTL Command decode support functions
PAGE
;************************************************************************
;*									*
;* parse_str_def-reads a command parameter and translates it to 	*
;*		 a sequence of characters. Parameter is of form 	*
;*									*
;*		<strn><termination>					*
;*	   where							*
;*		<strn> = {|<char>|<quoted_string>}[,<strn>]		*
;*		<char> = decimal number 				*
;*		<quoted_string> = '0 or more characters'                *
;*		<termination> = {:|<nul>}				*
;*									*
;*		 Sets carry if parse fails, clears if ok.		*
;*									*
;************************************************************************
; On entry
;
;   SI = Parameter string to parse
;   DI = Buffer to take output
;   Segment registers as start proc
;
; On exit
;
;   carry = true  ==> Parse failed
;     SI,DI maintained
;     CX undefined
;   carry = false ==> Parse ok
;     SI = pointer to termination character
;     DI = pointer to character after end translation
;     CX = number characters in translation
;   Other flags undefined
;   AX undefined
;   All other regs maintained.
;   Stack useage-18 bytes
;
;
parse_str_def proc near

  xor cx,cx
  push si
  push di				;Keep in case of need

  call stripspace
  jc psd01				;done

psd05:
  push cx
  call scan_quote			;Attempt to read as quoted string
  jnc psd02
  call scanint				;Attempt to read as int
  jnc psd06

					;Neither!
  pop cx
psd04:
  pop di
  pop si				;recover pointers
  stc					;return error
  jmp short psd03

psd06:					;Have found number
  pop cx
  or ah,ah
  jnz psd04				;Number to big
  mov [di],al				;store number
  inc di				;update pointer
  inc cx				;update count
  jmp short psd07

psd02:					;have found quoted string
  pop ax
  add cx,ax				;update count
psd07:
  call stripspace
  jc psd01				;Done-nul encountered
  cmp [si],byte ptr colon
  je psd01				;Done-':' encountered
  cmp [si],byte ptr comma		;seperator
  jne psd04				;bad exit
  inc si
  call stripspace
  jmp psd05				;get next <strn>

psd01:
  pop ax
  pop ax				;Drop old pointers
  clc					;return ok
psd03:
  ret
parse_str_def endp
;
;
;************************************************************************
;*									*
;*  scan_quote - scan asciiz string pointed to by SI for a quoted	*
;*		 string. Quote marks can be ' or ", and must            *
;*		 match. Sets carry if cannot build a valid quoted	*
;*		 string. Result is written to buffer at DI.		*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to suspected quoted string.
;   DI = pointer to location to write string
;   Segment registers as start proc
;
; On exit
;
;   Carry = true  ==> Not a quoted string
;     DI,SI maintained
;     CX undefined
;   Carry = false ==> Is a quoted string
;     SI = first character after final quote mark
;     DI = first available byte after representation of string
;     CX = number of characters in string
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-2 bytes
;
;
scan_quote proc near
  mov al,[si]
  cmp al,quotes1			;Look for quotation mark
  je scq01
  cmp al,quotes2
  je scq01
  jmp short scq05			;Not quote mark,fail

scq01:					;Found valid quotation mark
  push si				;save in case can't build string
  mov cx,-1
scq04:
  inc si
  cmp [si],byte ptr 0
  je scq03				;Nul found, exit bad
  inc cx
  cmp al,[si]				;Look for terminating quote
  jne scq04
					;Found terminating quote-
  pop si				;Copy over string
  push cx
  inc si				;First character of string
  rep movsb				;copy over string
  inc si				;skip over final quotation mark
  pop cx				;Number characters
  clc
  jmp short scq02			;return good

scq03:
  pop si				;restore SI
scq05:
  stc
scq02:					;done
  ret
scan_quote endp
;
;
;************************************************************************
;*									*
;*  scanint- scan asciiz string pointed to by SI for a decimal number	*
;*	     halts on non-digit or eol. If overflow (ie > 65535)	*
;*	     returns conversion of first 4/5				*
;*	     digits. If not valid number at all set carry.		*
;*									*
;************************************************************************
; On entry
;
;   SI = first character of number
;   Segment registers as start proc
;
; On exit
;
;   SI points to first non-digit character
;   carry = false ==> number
;     AX contains number
;   carry = true ==> not a number
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-10 bytes
;
;
scanint proc near
  push dx
  push cx
  push bx				;save regs
  push di

  xor bh,bh
  xor ax,ax
  mov di,10
  call scandigit
  jc sci01				;bad digit exit
sci02:
  mul di				;ax := ax * 10
  jc sci03				;overflow
  add ax,bx				;ax := ax + bl
  jc sci03				;overflow
  mov cx,ax				;store number
  call scandigit			;next digit
  jc sci04				;natural termination on non-digit
  jmp sci02

sci03:					;overflow-scan remaining digits
  call scandigit
  jnc sci03

sci04:
  mov ax,cx				;return number in ax
  dec si				;back up to recover non-digit
  clc					;Flag number
  jmp short sci05

sci01:
  dec si				;back up to recover non-digit
  stc					;flag not a number
sci05:
  pop di
  pop bx
  pop cx
  pop dx				;recover other regs
  ret
scanint endp
;
;
;************************************************************************
;*									*
;*  scandigit-ascii digit pointed to by SI is converted to number.	*
;*	      if not 0...9 then set carry. SI incremented.		*
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to digit
;   Segment registers as start proc
;
; On exit
;
;   SI byte following digit
;   carry = false
;     BL = value of digit
;   carry = true
;     BL ascii code
;   all other registers maintained
;   Other flags undefined
;   Stack useage-none
;
scandigit proc near
  mov bl,[si]
  cmp bl,'0'
  jb scd01				;fails
  cmp bl,'9'
  ja scd01				;fails
  sub bl,'0'                            ;calculate value
  clc
  jmp short scd02			;succede
scd01:
  stc
scd02:
  inc si				;move pointer on
  ret					;done
scandigit endp
;
;
;************************************************************************
;*									*
;*  Stripspace-advances SI past any space or tab characters. If meets	*
;*	       nul then sets carry flag else clears carry.		*
;*									*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   SI = pointer to string to be processed
;
; On exit
;
;   SI = pointer to first non-space character in string
;   carry = true  ==> nul encountered
;   carry = false ==> other character encountered
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-none
;
stripspace proc near
;
stps01: 				;Loop
  cmp [si],byte ptr spce		;look for space-continue loop
  je stps02
  cmp [si],byte ptr tab 		;look for tab-continue loop
  je stps02
  cmp [si],byte ptr nul 		;look for nul-exit and set carry
  je stps03
  jmp short stps05			;none of above-exit and clear carry

stps02:
  inc si
  jmp short stps01			;loop ends

stps03:
  stc					;nul encountered
  jmp short stps04

stps05: 				;Exit nul not found
  clc
stps04:
  ret
stripspace endp
;
;************************************************************************
;*									*
;*  Skiptospace-advances SI to first space or tab character. If meets	*
;*	       nul then sets carry flag else clears carry.		*
;*									*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   SI = pointer to string to be processed
;
; On exit
;
;   SI = pointer to first non-space character in string
;   carry = true  ==> nul encountered
;   carry = false ==> other character encountered
;   Other flags undefined
;   All other regs maintained.
;   Stack useage-none
;
skiptospace proc near
;
skts01: 				;Loop
  cmp [si],byte ptr spce		;look for space-quit loop
  je skts05
  cmp [si],byte ptr tab 		;look for tab-quit loop
  je skts05
  cmp [si],byte ptr nul 		;look for nul-quit and set carry
  je skts03
					;none of above-continue loop
  inc si
  jmp short skts01			;loop ends

skts03:
  stc					;nul encountered
  jmp short skts04

skts05: 				;Exit space/tab found
  clc
skts04:
  ret
skiptospace endp
;
;
;************************************************************************
;*									*
;*  Str_toupper - scan ASCIIZ string pointed to by SI, set all		*
;*		  lower case characters to upper case.			*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to string
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   All regs maintained
;   Stack useage-4 bytes
;
;
str_toupper proc near

  push si				;Save SI
  push ax				;Save AX

  xchg di,si
stu01:					;Convert all characters to upper case
  mov al,[di]
  or al,al
  jz stu02				;exit on zero
  cmp al,'a'                            ;Check is a...z
  jb stu03
  cmp al,'z'
  ja stu03
  and al,toupper			;Set to upper case
stu03:
  stosb
  jmp short stu01

stu02:
  xchg di,si
  pop ax
  pop si

  ret
str_toupper endp
;
;
;************************************************************************
;*									*
;*  Find_env-Search through environment for specified name. Returns	*
;*	     pointer to matching name, and count of characters. 	*
;*									*
;*									*
;************************************************************************
; On entry
;
;   ES:DI = pointer to environment
;   SI = pointer to name to match (Must include '=' sign)
;   Other segment registers as start proc
;
; On exit
;
;   Carry = false ==> matching name found
;     ES:DI = pointer to associated string
;     CX = length of string
;   Carry = true  ==> no match found
;     ES maintained
;     DI,CX undefined
;   Other flags undefined
;   AX,BX undefined
;   All other regs maintained
;   Stack useage-4 bytes.
;
;
find_env proc near
  push es
  push ds
  pop es
  push di				;save es:di
  mov di,si				;es:di := ds:si
  mov bx,di
  xor al,al
  mov cx,-1
  cld
  repne scasb				;look for nul
  dec di
  sub di,bx
  mov bx,di				;length of name
  pop di				;restore es:di
  pop es

fen01:
  cmp es:[di],byte ptr nul
  je fen02				;Done, no match found

  mov cx,bx				;byte count name
  push si
  repe cmpsb
  je fen04				;found matching name
  pop si
  mov cx,-1
  repne scasb				;skip past end of string
  jmp short fen01			;try next name

fen04:					;found matching name
  pop si
  mov bx,di
  mov cx,-1
  repne scasb				;find end string
  dec di
  sub di,bx
  mov cx,di				;length
  mov di,bx				;pointer to string
  clc
  jmp short fen03			;done

fen02:					;exit bad
  stc
fen03:
  ret
find_env endp
;
;
;************************************************************************
;*									*
;*  Skip_tostring- increment pointer past specified number ASCIIZ	*
;*		   strings.						*
;*									*
;************************************************************************
; On entry
;
;   AH = number of strings to skip
;   DI = pointer to first string
;   Segment registers as start proc
;
; On exit
;
;   DI = pointer to resulting string
;   Flags undefined
;   All other regs maintained
;   Stack useage-2 bytes
;
;
skip_tostring proc near
  push ax
  xor al,al
  mov cx,-1
  cld
sks01:
  or ah,ah
  jz sks02
  repne scasb				;look for nul
  dec ah
  jmp sks01				;next string

sks02:
  pop ax
  ret
skip_tostring endp
;
;
;************************************************************************
;*									*
;*  Str_cpy-copy string at SI to DI. Former is nul-terminated		*
;*									*
;*									*
;*									*
;************************************************************************
; On entry
;
;   SI = pointer to source string
;   DI = pointer to destination
;   Segment registers as start proc
;
; On exit
;
;   Flags undefined
;   SI,DI = bytes following respective strings
;   All other regs maintained
;   Stack useage-none
;
;
str_cpy proc near
  cld
stc01:
  lodsb 				;get byte
  or al,al				;Look for nul byte
  stosb 				;store it
  jnz stc01				;Again if not nul

  ret
str_cpy endp
;
;
SUBTTL Console input/output routines
PAGE
;************************************************************************
;*									*
;*  Kbdinp-routine reads character from keyboard (stdin). If CTRL-C	*
;*  has been pressed then int 23H is executed.				*
;*									*
;************************************************************************
;
; On entry at 'kbdnowait'
;   Other regs undefined
; On entry at 'kbdwait'
;   regs undefined
;
; On exit
;   carry = true ==> character has been read ( always so for entry at 'kbdwait')
;     AL = character read
;     AH undefined
;   carry = false ==> no character read
;     AL undefined
;     AH undefined
;
;   Other flag undefined
;   All other regs maintained.
;   Stack useage- As DOS services 0BH and 08
;
kbdnowait proc near		;Keyboard poll entry
  mov ah,0BH			;DOS service check keyboard status
  int dos
  or al,al
  jz kbd01			;exit if no characters in buffer

kbdwait label near		;Keyboard wait entry point
  mov ah,07H			;DOS service read keyboard
  int dos
  cmp al,cr			;Rerturn key
  jne kbd03
  mov al,newl			;Translate cr to newline (= lf)
kbd03:				;lf is used internally as end-of-line
				;character.

  stc				;Set carry, character read
  jmp short kbd02
kbd01:
  clc				;Clear carry, no characters read
kbd02:
  ret
kbdnowait endp
;
;************************************************************************
;*									*
;* wrint-write a 16-bit integer as an ascii string.			*
;*									*
;************************************************************************
; On entry
;
;   AX = source flag, one of fromctrl, fromecho, fromport
;   BX = number to convert
;   Segment registers as start proc
;
; On exit
;
;   AX undefined
;   Flags undefined
;   All other regs maintained.
;   Stack useage-12 bytes + DOS services
;
wrint proc near
  push bx
  push ax
  push di
  push dx

  mov ax,bx
  mov di,offset scrtmp + 6		;working buffer
  mov [di],word ptr 0			;nul-terminate
  mov bx,10

wri01:
  xor dx,dx
  div bx				;find least significant digit
  add dx,'0'                            ;find ascii
  dec di
  mov [di],dl				;store
  or ax,ax
  jnz wri01				;loop if not done

  mov bx,di				;pointer to characters
  pop dx
  pop di

  pop ax
  call screen

  pop bx
  ret
wrint endp
;
;
;************************************************************************
;*									*
;*  screenewl-calls screen with pointer to newline-nul sequence.	*
;*									*
;************************************************************************
; On entry
;
;   Segment registers as start proc
;   AX = source flag as for screen
;
; On exit
;
;   Flags maintained
;   All regs maintained.
;   Stack useage- 10 bytes + DOS services 02H
;
;
screenewl proc near
  push bx			;Save BX
  mov bl,newl
  call screenchar
  pop bx
  ret
screenewl endp
;
;
;************************************************************************
;*									*
;*  Screen-routine sends ASCIIZ string to screen (stdout).		*
;*  AX defines type of access-control info, echo keyboard, from port	*
;*									*
;************************************************************************
;
; On entry
;   AX = source flag, one of fromctrl, fromecho, fromport
;   BX = pointer to nul-terminated string to be output
;
; On exit
;
;   Flags maintained
;   All other regs maintained.
;   Stack useage- 8 bytes + DOS services 02H
;
;
  fromport EQU 0
  fromecho EQU 1
  fromctrl EQU 0
;
screen proc near
  pushf
  push ax			;Ignore source
  push bx
  push dx			;Save regs
scr01:
  mov dl,[bx]			;Get character to go
  or dl,dl
  jz scr02			;Terminate ouptut on nul character
  cmp dl, newl			;Check for newline character
  jne scr03
;output newline as cr,lf
  mov dl,cr
  mov ah,02H			;DOS display character
  int dos
  mov dl,lf

scr03:
  mov ah,02H			;DOS display character
  int dos
  inc bx
  jmp short scr01

scr02:
  pop dx
  pop bx
  pop ax
  popf				;Restore regs.
  ret
screen endp
;
;************************************************************************
;*									*
;*  Screenchar-sends character in bl to screen (stdout).		*
;*  AX defines type of access-control info, echo keyboard, from port	*
;*  (This service allows a nul to be sent to  stdout).			*
;*									*
;************************************************************************
;
; On entry
;   AX = source flag, one of fromctrl, fromecho, fromport
;   BL = Character to send
;
; On exit
;
;   Flags maintained
;   All other regs maintained.
;   Stack useage- 6 bytes + DOS services 02H
;
screenchar proc near
  pushf
  push ax			;Ignore source
  push bx			;Save regs
  xchg bx,dx
scc01:
  cmp dl, newl			;Check for newline character
  jne scc03
;output newline as cr,lf
  mov dl,cr
  mov ah,02H			;DOS display character
  int dos
  mov dl,lf

scc03:
  mov ah,02H			;DOS display character
  int dos

  xchg bx,dx
  pop bx
  pop ax
  popf				;Restore regs.
  ret
screenchar endp
;
;
;
;
  code_size equ $			;size of code segment
code ends


end start
