/************************************************************************/
/*									*/
/*		Simple terminal program for MS-DOS with 		*/
/*              Serial Solutions Asynchronous port driver.              */
/*									*/
/*         Written by Richard Davis, 22/4/90, Copyright (c) 1990        */
/*									*/
/************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <fcntl.h>
#include <io.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <time.h>

#define TRUE -1
#define FALSE 0

enum SCREEN {		  /* Defines the output screens available to */
	      fromport,   /* program. One for characters from port. */
	      fromecho,   /* Another for echo from keyboard.	    */
	      fromctrl	  /* Another for error messages, menus, etc. */
	    };

/* Define default settings.                      */

int comno = 1;               /* COM1      */
char comname[] = "COM1\0   "; /* Name- might not be a com port. */
long baud = 1200l;       /* Baud rate */
char parity = 'e';        /* Parity    */
int databits = 7;       /* Data bits */
int stopbits = 1;       /* Stop bits */
int brk = FALSE;        /* Break on serial line. */

enum ECHO { echon, echos, echol, echols }; /* Echo off, screen, line, both*/
enum ECHO echo = echon;        /* Echo off   */

char def_file[]   = "termdef.txt"; /* File holding terminal definitions. */

double timeout = 5e0;           /* Five secend timeout waiting for */
				/* o/p buffer                      */
int writefail = FALSE;          /* Flag for write failures. */
char *gppath;

FILE *com_inp,  /* Files for: com port input */
     *com_out;  /*           com port output */

void main(int argc, char **argv); /* Main Program */

#define XLT_SIZE 1024
typedef char *xlt[257]; 	 /* translation table. */
xlt inxlt;	  /* Translation table for inxlat. */
xlt outxlt;	  /* Translation table for outxlat. */

void inxlat(unsigned char *buffer);  /* Translate i/p from port */
void outxlat(unsigned char *data);     /* Translate o/p to port */

int inpstr(char *data);             /* Receive string from port. */
void outstr(char *data, int count); /* Send string to port.      */


void init_xlat(void);      /* Initialise translation tables. */
void init_screen(void);    /* Initialise screen. */
void exec_defaults(char **argv);           /* Find and execute a terminal */
					   /* definition file.   */
void open_com(void);		/* Open com port. */


int kbdinp(char *chr, int kb_retry);    /* Read input from keyboard */
void screen(char *output, enum SCREEN scr); /* Send data to screen(s). */
char scrbuff[256];                      /* character o/p buffer */

void switchout(void);                   /* Read from keyboard, process */
					/* if command or send out to port.*/
void menuout(void);                     /* Process menu or command */
void cmdout(char *cmd);                 /* Process command. */
void readin(void);      /* Read from port and send to screen. */

/* Command functions. */
void setup(void);                       /* Set up port. */
int sumenu( char *title, char** options); /* Menu function. */
int sumain(void);                         /* Main menu function. */

void set_xlt(xlt xtable, char *cmdstr); /* Set translation. */
int set_xlt_low(xlt xtable, char *key,               /* Set translation */
		char *keyend, char *rep, char *repend);

void help(void);                         /* Display Help information.*/
void set_echo(char *ech);                /* Set echo type. */
void set_com(char *comparam);            /* Set Com Port.            */
void set_baud(char *param);              /* Set Baud rate.           */
void set_parity(char *param);            /* Set parity.              */
void set_databits(char *param);          /* Set number data bits.    */
void set_stopbits(char *param);          /* Set number stop bits.    */
void get_file(char *param);              /* Get file from port */
int quit_gfile( FILE *gfile);            /* Check for user halt */
void put_file(char *param);              /* send file to port. */

int bios_x_init(int port, int brk, char par, int stop, int data, long baud);
					/* Set parameters. */
int prefer_baud(long *baud);            /* Check baud rate. */

void exec_file(char *name);              /* Execute command file. */
void exec_dos(char *cmd);                /* Execute DOS command. */

void transmit(char *seq);               /* Send a sequence of characters */

  /* General support routines. */
void uncook(FILE *file);    /* Change file from cooked to raw data */
void str_toupper(char* str);            /* Set string to upper case. */
int parse_str_def(char **str, char **opseq); /* Build parameter string. */
void drop_space(char **str);         /* Strip whits space from string. */
int is_term(char ch);                   /* Return True if terminator. */
int is_quote(char ch);                  /* Return true if quote mark. */

void main(int argc, char **argv)
{
  int dummy;

  init_screen();                        /* Set up screen. */
  init_xlat();                          /* initialise translation tables. */
					/* This may need to be replaced. */

  exec_defaults(argv);           /* Find and execute a file of commands */

  open_com();			       /* Open com files. */

  while( TRUE ) 			  /* Main program loop. */
  {
      readin();             /* get any characters from port. */
			    /* send to display.              */

      switchout();          /* Read and process characters */
			    /* from keyboard. */
  };
}

  /* Read in and translate one or more characters. */
  /* The translation is written to 'buffer' as a nul-terminated string. */
void inxlat(unsigned char *buffer)
{
  int n;
  char *xpt, *xptend;

  n = inpstr(buffer);             /* Read in character. */

  /* inxlt['a'] is a pointer to the string which replaces 'a'.  */
  /* The string is not nul-terminated, but ends at the begining */
  /* of the string for 'b', pointed to by inxlt['b'].           */

  if (n != 0)                        /* If have read a character */
  {
    xptend = inxlt[(*buffer)+1];     /* Find end of buffer. */
    for(
	xpt = inxlt[*buffer];        /* Find start of translation */
	xpt != xptend;               /* Test for end of translation */
	buffer++, xpt++
       )
	*buffer = *xpt;              /* copy over */
  };
  *buffer = '\0';                    /* Terminate string */
}

  /* Read a character from com port. */
  /* If echo to line true then send it back out. */
  /* Return value is number of bytes read.       */
int inpstr(char *data)
{
  int c;
  c = getc(com_inp);
  *data = (char)c;

  if( c == EOF )
  {
    clearerr(com_inp);
    return 0;            /* EOF is error condition ==> return none. */
  }
  else
  {
    if( echo & echol )    /* If echo set then send back out. */
      outstr(data, 1);
    return 1;
  };
}

  /* Translate a (some) characters and output them. */
  /* data is a nul-terminated string to be output. */
void outxlat(unsigned char *data)
{
  int count;              /* Number of characters in each translation. */
  unsigned char *dpt = data;       /* Pointer to current character to send. */

  while( *dpt )    /* For all characters in string. */
  {
    /* outxlt['a'] is a pointer to the string which replaces 'a'. */
    /* The string is not nul-terminated, but ends at the begining */
    /* of the string for 'b', pointed to by outxlt['b']. Hence a  */
    /* nul can be included in an output translation.              */

    count = outxlt[(*dpt)+1] - outxlt[*dpt]; /* Number characters */

    if(count)                    /* Can translate to no characters at all */
      outstr(outxlt[*dpt],count);  /* Send one or more characters. */

    dpt++;                                /* Next character. */
  };
}

  /* Send a string  of characters to com port. */
  /* This routine must deal with write errors on COM port. */
void outstr(char *data, int count)
{
  int i,c;
  time_t t1, t2;

  for(i = 0; i < count ; i++)
  {
    time( &t1 );
    fputc(*(data + i), com_out);
    fseek(com_out, 0l, SEEK_SET);  /* Reset file pointer..means flush  */
    while( ferror(com_out) && (difftime(time(&t2),t1) < timeout) )
    {
      clearerr(com_out);
      fputc(*(data + i), com_out);
      fseek(com_out, 0l, SEEK_SET);  /* Reset file pointer..means flush  */
    };
    if( ferror(com_out) )
    {
      screen("Failed to output characters to port.\n",fromctrl);
		       /* Write failure */
      clearerr(com_out);
    };
  };
}

  /* Function reads characters from keyboard. */
  /* Processes any commands and sends all other bytes on to port.*/
void switchout(void)
{
  int chr;
  char c[2];

  *(c+1) = '\0';
  chr = kbdinp(c, FALSE);     /* Read in a character. */
  if(chr)               /* If have read any. */
  {
    if( *c == '\0' )      /* Trap command/menu code. */
      menuout();               /* Process as menu command. */
    else
    {
      if( echo & echos ) /* May need to echo character to screen. */
	screen(c,fromecho);
      outxlat(c);     /* Send character on to port. */
    };
  };
}

/* Read from port and send to screen. */
void readin(void)
{
  char bf[50];

  inxlat(bf);   /* Read in some characters. */
  if(*bf != '\0' )   /* If some read in ok. */
    screen(bf,fromport);
}

  /* Process menu key function. */
  /* Unrecognised keys are dropped. */
  /* (char)46 is passed on to command processor (is ALT-C). */
  /* Menu keys are used to initiate menus, which in turn generate output */
  /* sent to command processor. */
void menuout(void)
{
  int chr, done;
  char c;
  char cmd[120];        /* Buffer for commands. */

  chr = kbdinp(&c, FALSE);     /* Read in a character. */
  if(chr)
  {
    switch( c )
    {
      case 46:  /* ALT-C reads in a command */
		screen("Type command:",fromctrl);
		scanf(" %119[^\n]",cmd);
		cmdout(cmd);
		break;
      case 59:  /* F1: Help option. */
		cmdout("h");
		break;
      case 60:  /* F2: Setup option. */
		setup();
		break;
      case 68:  /* F10: Quit option. */
		cmdout("q");
		break;
      default:  /* Unrecognised extended code is dropped. */
		break;
    }
  };
}

/* Process command. */
/* Commands are single characters (case ignored), followed by optional */
/* parameters. They are terminated by an end_of_input or by cr/lf.     */
void cmdout(char *cmd)
{
  char tmp[50];               /* Temporary buffer.     */

  switch( tolower(*cmd) )        /* Decode single character command. */
  {
    case 'q':  /* Quit terminal command. */
	       screen("Quitting program.\n",fromctrl);
	       fclose(com_inp);  /* Tidy up. */
	       fclose(com_out);
	       exit(0);         /* Will halt program. */
	       break;
    case 'i':  /* Set input translation. */
	       screen("Setting input translation\n", fromctrl);
	       set_xlt(inxlt,cmd+1);
	       break;
    case 'o':  /* Set output translation. */
	       screen("Setting output translation\n", fromctrl);
	       set_xlt(outxlt,cmd+1);
	       break;
    case 'h':  /* Display Help. */
	       help();
	       break;
    case 'c':  /* Set com port to use. */
	       set_com(cmd+1);
	       break;
    case 'b':  /* Set baud rate. */
	       set_baud(cmd+1);
	       break;
    case 'p':  /* Set parity. */
	       set_parity(cmd+1);
	       break;
    case 'd':  /* Set data bits. */
	       set_databits(cmd+1);
	       break;
    case 's':  /* Set stop bits. */
	       set_stopbits(cmd+1);
	       break;
    case 'x':  /* Execute file as command. */
	       exec_file(cmd+1);
	       break;
    case 'r':  /* Execute dos command. */
	       exec_dos(cmd+1);
	       break;
    case 'e':  /* Set echo */
	       set_echo(cmd+1);
	       break;
    case 't':  /* Transmit sequence */
	       transmit(cmd+1);
	       break;
    case '\'': /* Remark. Used for command files. */
	       screen(cmd+1, fromctrl); /* echo line. */
	       screen("\n", fromctrl);
	       break;
    case 'g' : /* Get file. used to copy data from serial port to file*/
	       get_file(cmd+1);
	       break;
    case 'u' : /* Put file. Used to send a file to serial port. */
	       put_file(cmd+1);
	       break;
    case '\n':
    case '\l':
    case '\0':
	       /* No action. */
	       break;
    default:   /* Unrecognised command. */
	       screen("Command not recognised\n",fromctrl);
	       break;
  };
}

  /* Set up port. */
void setup(void)
{
  /* Menu titles and options for set. */
  static char *portmn[]   = {"COM1", "COM2", "COM3", "COM4", "COM5",
			     "COM6", "COM7", "COM8", "Leave alone", NULL};
  static char *baudmn[]   = {"110 Baud", "150 Baud", "300 Baud",
			     "600 Baud", "1200 Baud", "2400 Baud",
			     "4800 Baud", "9600 Baud", "Leave alone", NULL};
  static char *parmn[]    = {"None", "Odd", "Even", "Mark", "Space",
			     "Leave alone", NULL};
  static char *dabmn[]    = {"5 Bits", "6 Bits", "7 Bits", "8 Bits",
			     "Leave alone", NULL};
  static char *stomn[]    = {"1 Bit", "2 Bits", "Leave alone", NULL};

  char titles[80];
  int mchoice;

  while( TRUE )
  {
    mchoice = sumain();  /* Main menu...list all current settings. */
    switch( mchoice )
    {
      case 0: /* Set com port. */
	      sprintf(titles,
		      "  Select Port To Use (now %s)",
		      comname);
	      mchoice = sumenu(titles, portmn);      /* Get port. */
	      if( mchoice != 8 )                     /* 8 is leave alone */
		set_com( portmn[mchoice] );          /* Set it.   */
	      break;
      case 1: /* Set Baud rate. */
	      sprintf(titles,
		      "  Select Baud Rate (now %li)",
		      baud);
	      mchoice = sumenu(titles, baudmn);      /* Get baud rate. */
	      if( mchoice != 8 )                     /* 8 is leave alone */
		set_baud( baudmn[mchoice] );               /* Set it.   */
	      break;
      case 2: /* Set Parity. */
	      sprintf(titles,
		      "  Select Parity (now %s)",
		      parity == 'n' ? "none" :
		      parity == 'e' ? "even" :
		      parity == 'o' ? "odd" :
		      parity == 'm' ? "mark" : "space"
		      );
	      mchoice = sumenu(titles, parmn);      /* Get baud rate. */
	      if( mchoice != 5 )                     /* 5 is leave alone */
		set_parity( parmn[mchoice] );               /* Set it.   */
	      break;
      case 3: /* Set databits. */
	      sprintf(titles,
		      "  Select Number Data Bits (now %i)",
		      databits);
	      mchoice = sumenu(titles, dabmn);       /* Get data bits. */
	      if( mchoice != 4 )                     /* 4 is leave alone */
		set_databits( dabmn[mchoice] );      /* Set it.   */
	      break;
      case 4: /* Set stop bits. */
	      sprintf(titles,
		      "  Select Number Stop Bits (now %i)",
		      stopbits);
	      mchoice = sumenu(titles, stomn);       /* Get stop bits. */
	      if( mchoice != 2 )                     /* 2 is leave alone */
		set_stopbits( stomn[mchoice] );      /* Set it.   */
	      break;
      case 5: /* Continue terminal. */
	      screen("Setup Complete. Returning to terminal session.\n",
		     fromctrl);
	      return;
	      break;
      case 6: /* Quit. */
	      cmdout("q");
	      break;
    };
  };
}

  /* Main setup menu. Displays list of current settings and asks */
  /* user to change/continue. Return value is selection.         */
int sumain(void)
{
  static char mainname[] =
  "===========Setup Main Menu.==========\n      Alter terminal settings.\7";
  char *mn[8];              /* Holds options.   */
  char buffer[160];         /* Holds Menu text. */
  char *count;

    /* Define menu text- uses current port settings. */
  mn[0] = buffer;
  mn[1] = mn[0] + 1 + sprintf(mn[0], "Port (now %s)", comname);
  mn[2] = mn[1] + 1 + sprintf(mn[1], "Baud rate (now %li)", baud);
  mn[3] = mn[2] + 1 + sprintf(mn[2],"Parity (now %s)",
					parity == 'n' ? "none" :
					parity == 'e' ? "even" :
					parity == 'o' ? "odd" :
					parity == 'm' ? "mark" : "space"
			     );
  mn[4] = mn[3] + 1 + sprintf(mn[3], "No. data bits (now %i)", databits);
  mn[5] = mn[4] + 1 + sprintf(mn[4], "No. stop bits (now %i)", stopbits);
  mn[6] = mn[5] + 1 + sprintf(mn[5], "Continue terminal");
		      sprintf(mn[6], "Quit terminal");
  mn[7] = NULL;

  /* Call menu function. */
  return sumenu(mainname, mn);
}

  /* Menu function. Operands are title and list of options. */
  /* Returns user selection. */
int sumenu( char *title, char** options)
{
  char bf[80];
  int num, sel;
  char c;

  screen("\n\n", fromctrl);             /* Print out menu title. */
  screen(title, fromctrl);
  screen("\n\n", fromctrl);

  screen("  Key  For....\n", fromctrl);
  num = 0;
  while( *options != NULL )      /* Print out options. */
  {
    sprintf(bf, "  <%c>  %s\n", '1' + num++, *options++);
    screen(bf, fromctrl);
  };

  sprintf(bf,"\n select option :");
  screen(bf, fromctrl);

  do
  {
    kbdinp(&c, TRUE);            /* Get a character from keyboard. */
    sel = c - '1';              /* Turn into option number. */
  }
  while
  (
    sel < 0 || sel >= num       /* Repeat if not valid selection */
  );
  sprintf(bf, "%c\n", c);          /* Print selection */
  screen(bf, fromctrl);

  return sel;
}

/* Get file from port. Characters read from serial port are  */
/* copied to file specified. File is closed on time-out */
/* (five seconds), or on receipt of optional termination character */
/* Incoming characters are checked against terminal sequence     */
/* after latter has been translated.                            */
/* Format of operands is:                                       */
/*        <pathname>[n][ <term_sequence>]                       */
/* <term_sequence> has same form as sequences in translate and  */
/* transmit commands. The n switch forces suppression of translations */
void get_file(char *param)
{
  FILE *gfile;
  char c[2];
  int ch,termlen;
  char term[80];
  char *tmp,*termend;
  time_t t1, t2;
  int end_of_gfile;
  char buffer[256];
  long flen;
  int notranslate;

  drop_space(&param);
  gppath = param;
  while( *param && !isspace(*param) )
    param++;
  if( isspace(*param) )
    *(param++) = '\0';

  gfile = fopen(gppath, "r");
  if( gfile != NULL )
  {
    screen("File already exists overwrite, (hit 'y' or 'n')?",
	   fromctrl);
    kbdinp(c,TRUE);
    screen("\n",fromctrl);
    if( tolower( *c ) != 'y' )
    {
      fclose( gfile );
      screen("Returning to terminal.\n",fromctrl);
      return;
    };
  };

  gfile = fopen(gppath, "wb");
  if( gfile == NULL )
  {
    sprintf(scrbuff, "Cannot open file %s. Returning to terminal.\n",gppath);
    screen(scrbuff,fromctrl);
    return;
  };

  drop_space(&param);
  if( tolower(*param) == 'n' )
    notranslate = TRUE;
  else
    notranslate = FALSE;
  param++;
  termend = term;
  if( !parse_str_def(&param, &termend) )
  {
    screen("Cannot understand termination. Returning to Terminal.\n",
	   fromctrl);
    fclose( gfile );
    return;
  };
  *termend = '\0';      /* nul-terminate string*/

  screen("Reading characters into file. Hit any key to pause.\n",fromctrl);
  screen("        bytes read into file.",fromctrl);

  flen = 0l;
  ch = 0;
  while( TRUE )
  {
    end_of_gfile = FALSE;
    if( termend == term )
    {
      /* Get characters on timeout */
      time(&t1);     /* Start clock */
      while( TRUE )
      {
	if( notranslate )
	  ch = inpstr(buffer);
	else
	{
	  inxlat(buffer);
	  ch = strlen(buffer);
	};
	if( ch )   /* exit timing loop if some read */
	  break;
	if( difftime( time(&t2), t1 ) > timeout )
	{
	  end_of_gfile = TRUE;
	  break;
	};
	if( quit_gfile(gfile) )
	  return;
      };
    }
    else
    { /* Get characters until termination sequence. */
      while( TRUE )
      {
	if( notranslate )
	  ch = inpstr(buffer);
	else
	{
	  inxlat(buffer);
	  ch = strlen(buffer);
	};
	if( ch )
	  break;
	if( quit_gfile(gfile) )
	  return;
      };
      tmp = buffer;
      end_of_gfile = FALSE;
      for( tmp = buffer; tmp < buffer + ch; tmp++ )
	if( *tmp == *term )
	{  /* Read end_of_file marker */
	  ch = tmp - buffer;
	  end_of_gfile = TRUE;
	  break;
	};
    };
    for( tmp = buffer; tmp < buffer + ch; tmp++ )
      if( fputc(*tmp, gfile) == EOF )
      {
	sprintf(scrbuff,"Error on write to file %s.\n",gppath);
	screen(scrbuff,fromctrl);
      };

    flen += ch;
    sprintf(scrbuff,"\r%li",flen);
    screen(scrbuff,fromctrl);

    if( end_of_gfile )
    {
      fclose(gfile );
      screen("\nFinished writting to file.\n",fromctrl);
      screen("Returning to terminal.\n",fromctrl);
      return;
    };
  };
}

/* Quit_gfile- check for a keypress- if true then asks user whether or */
/* not to continue reading in file. If user wants to quit then close file */
int quit_gfile( FILE *gfile)
{
  char c;
  if( kbdinp(&c,FALSE) )
  {
    screen("Quit receiving file (hit 'y'), or continue (hit any other key)",
	    fromctrl);
    kbdinp(&c,TRUE);
    screen("\n",fromctrl);
    if( tolower( c ) == 'y' )
    {
      fclose( gfile );
      remove(gppath);
      screen("Returning to terminal\n",fromctrl);
      return TRUE;
    };
  };
  return FALSE;
}

/* send file to port. File specified is opened, and sent to serial */
/* port after translation. Optional termination sequence is sent at */
/* end of file. Format is        */
/*       <pathname>[n][ <term_sequence>]                            */
/* <term_sequence> is of same format as for transmit and translate  */
/* commands. n switch is present to suppress translations.          */
void put_file(char *param)
{
  FILE *pfile;          /*destination file*/
  char *path;
  char c[2];
  int ch;
  char term[80];
  char *tmp,*termend;
  int notranslate;
  long flen;

  drop_space(&param);
  path = param;
  while( *param && (*param != ' '))
    param++;
  if( *param )
    *(param++) = '\0';

  pfile = fopen(path,"rb");
  if( !pfile )
  {
    sprintf(scrbuff,"Could not open file %s, returning to terminal\n",path);
    screen(scrbuff, fromctrl);
    return;
  };
  /* look for more parameters */
  drop_space(&param);
  if( tolower(*param) == 'n' )
    notranslate = TRUE;
  else
    notranslate = FALSE;
  param++;

  termend = term;
  if( !parse_str_def( &param, &termend ) )
  {
    screen("Could not understand the termination characters.\n",fromctrl);
    screen("Returning to terminal.\n",fromctrl);
    fclose(pfile);
    return;
  };
  flen = 0l;
  screen("Sending file. Hit any key to pause.\n",fromctrl);
  screen("        bytes sent from file.",fromctrl);
  ch = getc(pfile);
  while( !feof(pfile) )
  {
    c[0] = (char)ch;
    c[1] = '\0';
    if( notranslate )
      outstr(c,1);
    else
      outxlat(c);
    if( writefail )
    {
      screen("Returning to terminal\n",fromctrl);
      fclose( pfile );
      return;
    }
    else
    {
      flen += 1;
      sprintf(scrbuff,"\r%li",flen);
      screen(scrbuff,fromctrl);
    };
    if( kbdinp(c, FALSE) )
    {
      screen("Quit sending file (hit 'y'), or continue (hit any other key)",
	     fromctrl);
      kbdinp(c, TRUE);
      screen("\n",fromctrl);
      if( tolower( *c ) == 'y' )
      {
	screen("Returning to terminal\n",fromctrl);
	fclose( pfile );
	return;
      };
    };
    ch = getc(pfile);
  };
  fclose( pfile );

  /*  Send termination sequence */

  if( notranslate )
    outstr(term,termend - term);
  else
  {
    *termend = '\0';
    outxlat(term);
  };

  sprintf(scrbuff,"Done putting file %s, returning to terminal.\n",path);
  screen(scrbuff,fromctrl);
}

/* Set translation. */
/* Xtable is the xtranslation structure to set, */
/* cmdstr is a string containing the parameters to use. */
/* Return value is succede flag. FALSE halts program. */
/* Format is as follows:					*/
/* <string_def1>:<string_def2>					  */
/* where <string_def1>={<char_def>|<quoted_string1>}[,<string_def2>] */
/*	 <string2_def>={|{<char_def>|<quoted_string2>}[,<string_def2>]}  */
/*	 <char_def>=decimal 0..255					*/
/*	 <quoted_string1>="1 or more characters"                         */
/*	 <quoted_string2>={""|<quoted_string1>}                          */
void set_xlt(xlt xtable, char *cmdstr)
{
  char buff[512];	  /* Used to hold key and replacement strings. */
  char *begin, *key, *rep, *end;   /* Pointers into buff. */
  char *first, *second;            /* Pointers int command string. */

  rep = key = buff;

  first = cmdstr;
  if(
    !parse_str_def(&cmdstr, &rep)   /* Parse <string_def1> */
    )
  {
    screen("Cannot understand key string.\n",fromctrl);
    return;                         /* Fails */
  };

  if( key == rep )
  {
    screen("Key string must not be empty.\n",fromctrl);
    return;                         /* Fails */
  };
  if( *cmdstr != ':' )
  {
    screen("Strings must be separated by ':'.\n",fromctrl);
    return;                         /* Fails */
  };

  *cmdstr = '\0';
  second = ++cmdstr;
  end = rep;
  if(
    !parse_str_def(&cmdstr, &end)   /* Parse <string_def1> */
    )
  {
    screen("Cannot understand replacement string.\n",fromctrl);
    return;                         /* Fails */
  };

  screen("From ", fromctrl);            /* Describe translation. */
  screen(first, fromctrl);
  screen(" to ", fromctrl);
  screen(second, fromctrl);
  screen("\n", fromctrl);

  set_xlt_low(xtable, key, rep, rep, end);
}

  /*  write to translation tables the replacement for the key. */
  /* Returns True if succedes. */
  /* Note that strings are not nul-terminated */
int set_xlt_low(xlt xtable, char *key, char *keyend, char *rep, char *repend)
{
  int dlength;			/* Change in length. */
  int i;
  char *ptr;			/* General pointer */

  if( keyend != key + 1 )	/* Can translate from single char only. */
  {
    screen("Cannot make that transation, sorry.\n", fromctrl);
    return FALSE;
  };

  dlength = repend - rep -	(xtable[(*key)+1] - xtable[*key]);
       /* Length of new rep. */ /*   Length of old rep.        */

  if( ( xtable[257] - xtable[0] +	dlength      ) > XLT_SIZE )
      /* Length of table   */	  /* Change in size*/	 /*total*/
  {
    screen("Not enough space for that translation, sorry.\n", fromctrl);
    return FALSE;
  };

  if( dlength != 0 )	/* Need to change size of table. */
  {
    if( dlength > 0 )	  /* Need to shuffle table up to make space. */
    {
      for(ptr = xtable[256] - 1; ptr >= xtable[(*key)+1]; ptr--)
	*(ptr+dlength) = *ptr;
    }
    else		  /* Need to shuffle table backwards. */
    {
      for(ptr = xtable[(*key)+1]; ptr < xtable[256]; ptr++)
	*(ptr+dlength) = *ptr;
    };
    for(i = (*key) + 1; i <= 256; i++)
      xtable[i] += dlength;	    /* Alter pointers. */
  };

  /* Write in new representation. */
  for(ptr = xtable[*key]; rep != repend;  ptr++, rep++)
    *ptr = *rep;

  return TRUE;			/* Done */
}

  /* Read through a command string parameter and build a sequence   */
  /* of characters from it. String is of form <string_def2>, above. */
  /* Termination is ':', cr, lf or '\0'                             */
  /* Returns flag for successful parse. 			    */
int parse_str_def(char **str, char **opseq)
{
  char thisquote;
  int num;

  drop_space(str);                     /* Strip off white space. */

  while( ! is_term(**str) )
  {
    drop_space(str);			 /* Strip off white space. */

    if( is_quote(**str) )		 /* Look for quoted string. */
    {
      thisquote = *((*str)++);		 /* Remember starting character. */
      while( (**str != thisquote) && **str )
	*((*opseq)++) = *((*str)++);	   /* Copy over character. */
      if( !(**str) )
	return FALSE;			  /* Must have closing quote */
      (*str)++; 			  /* Drop quote. */
      drop_space(str);
      if( !is_term(**str) && (**str != ',') )
	return FALSE;			/* Bad string specification. */
    }
    else if( isdigit(**str) )		  /* Look for decimal number. */
    {
      num = 0;
      do
	num = num*10 + ( (int)*((*str)++) - (int)'0' );
      while( isdigit(**str) && num < 256);  /* 0<=num<=255 */
      drop_space(str);
      if( (!is_term(**str)) && (**str != ',') && (!isspace(**str) )
	  || (num > 255) )
	return FALSE;			/* Bad number specification. */
      *((*opseq)++) = (char)num;	 /* Copy over character. */
    }
    else
      return FALSE;			/* Cannot decode. */

    drop_space(str);
    if( **str == ',' )
      (*str)++;
  };
}

  /* Strips off white space characters. */
void drop_space(char **str)
{
    if( isspace(**str) && !is_term(**str) )
      (*str)++;     /* Strip off any white space */
}

  /* Returns TRUE if ch is ':', cr, lf or '\0'. */
int is_term(char ch)
{
  return( !ch || (ch == ':') || (ch == '\n') || (ch == '\r') );
}

  /* Returns TRUE if ch is '\'' or '\"'  */
int is_quote(char ch)
{
  return( (ch == '\'') || (ch == '\"') );
}

  /* Send a sequence of characters                           */
  /* Characters are specified as                             */
  /*       <seq>={<null>|<char_def>|<quoted_string>}[,<seq>] */
  /*       <char_def>=decimal 0..255                         */
  /*       <quoted_string>="0 or more characters"            */
  /*       <null> =                                          */
  /* The characters are not translated.                      */
void transmit(char *seq)
{
  char op[100];         /* Characters to output. */
  char *opend;

  opend = op;           /* End of list of characters */

  parse_str_def(&seq, &opend);

  outstr(op, opend - op);  /* Send string. Is not translated. */
}


  /* Display Help information. */
void help(void)
{
static char *helpinfo[] = {
"Function keys:\n",
" F1             help                   F10             quit\n",
" F2             Setup\n",
"Type ALT-C (ie press 'C' key while 'alt' key is held done) to issue command.\n",
"Commnads are:\n",
" H              help                   Q               quit\n",
" E{n|s|l|ls}    set echo screen, line  C<portname>     set port (e.g. COM2)\n",
" B<Baudrate>    set Baud rate          P<parity>       set parity {n|o|e|m|s}\n",
" D<databits>    set number data bits   S<stopbits>     set number stop bits\n",
" X<filename>   execute file            R<command...>   run DOS command\n",
" T<strn>       transmit sequence       '             comment\n",
" I<char>:<strn> translate input        O<char>:<strn>  translate output\n",
" U <filename>[ n][<char>]    put file <filename>, if 'n' no translation,\n",
" G <filename>[ n][<char>]    get file <char> character to terminate\n",
" ( <char> is a decimal number, or a character in quotation marks.\n",
"   <strn> is sequence of numbers or quoted strings separated by commas, Eg\n",
"   o12:12,\"New page\",13,10,10                                        )\n",
		       NULL
			      };
  char **line;

  line = helpinfo;
  while( *line != NULL )
    screen(*(line++), fromctrl);
}

  /* Set echo. */
  /* Echo can be none, to screen, to line or both. */
  /* These are "" or "n", "s",      "l"    "ls"    */
void set_echo(char *ech)
{
  echo = echon;
  while( *ech )
  {
    switch( tolower( *ech ) )
    {
      case 'n': echo = echon; /* No echo */
		break;
      case 'l': echo |= echol; /* Line echo */
		break;
      case 's': echo |= echos; /* Screen echo */
		break;
      case ' ': /* Ignore spaces */
		break;
      default:  screen("Unrecognised character in echo setting. Ignored.\n",
		       fromctrl);
		break;
    };
    ech++;
  };

  switch( echo )
  {
    case echon: /* No echo */
		screen("Echo off.\n",fromctrl);
		break;
    case echos: /* Echo to screen only */
		screen("Echo to screen only.\n",fromctrl);
		break;
    case echol: /* Echo to line only */
		screen("Echo to line only.\n",fromctrl);
		break;
    case echols: /* Both echo */
		screen("Echo to screen and line.\n",fromctrl);
		break;
  }
}

/* Set Com Port.            */
/* Parameter is string eq "COM2", if of form "COMn" n is stripped off */
/* and written to comno.   */
void set_com(char *comparam)
{
  int conv, com;
  char buff[60];

  str_toupper( comparam );                      /* Set to upper case. */

  if(
     sscanf(comparam, "%s", buff) == 1             /* Can get name. */
    )
  {
  if( com_inp != NULL )                          /* If has been opened */
  {
    fclose(com_inp);                               /* Close */
    fclose(com_out);
  };

  /* Look for a standard com port. */
  conv = sscanf(comparam, " COM%i ", &com );
  if( (com > 0) && (com <= 16) && (conv == 1) ) /* Is of correct format */
    comno = com;
  else
    comno = 0;                                  /* Not a COM port. */

  strncpy(comname, buff, 8);          /* Use (first 8 chars of) name */

  open_com();                           /* Open port. */

  /* Give some information. */
  sprintf(buff, "Port is:%s\n", comname);
  screen(buff, fromctrl);
  };
}

/* Set Baud rate.           */
void set_baud(char *param)
{
  long bd;
  int bprm;
  char bf[60];

  if( sscanf(param, " %li", &bd) == 1 )  /* Get in baud rate. */
  {
    if(
      bios_x_init(comno, brk, parity, stopbits, databits, bd)  /* Set */
      )
    {
      /* Setting o.k. */
      prefer_baud( &bd );        /* Expand an abreviation */
      baud = bd;                /* Save new baud rate. */
      sprintf(bf, "Set Baud rate at %li.\n", baud);
      screen(bf, fromctrl );
      return;              /* Done */
    };
  };

  screen("Cannot set that baud rate.\n", fromctrl);
}

/* Set parity.              */
void set_parity(char *param)
{
  char p;
  char bf[50];

  if(
     sscanf(param, " %c", &p) == 1
    )
  {
    if(
      bios_x_init(comno, brk, p, stopbits, databits, baud)  /* Set */
      )
    {
      /* Setting o.k. */
      p = tolower(p);
      parity = p;                /* Save new parity. */
      sprintf(bf, "Set %s parity.\n", ( p == 'n' ? "no" :(
					p == 'o' ? "odd" :(
					p == 'e' ? "even" :(
					p == 'm' ? "mark" :
						   "space" )))
				      )                         );
      screen(bf, fromctrl );
      return;              /* Done */
    };
  };

  screen("Cannot set that parity.\n", fromctrl);
}

/* Set number data bits.    */
void set_databits(char *param)
{
  int dab;
  char bf[50];

  if(
     sscanf(param, " %i", &dab) == 1
    )
  {
    if(
      bios_x_init(comno, brk, parity, stopbits, dab, baud)  /* Set */
      )
    {
      /* Setting o.k. */
      databits = dab;                /* Save new data bits. */
      sprintf(bf, "Set %i data bits.\n",  dab);
      screen(bf, fromctrl );
      return;              /* Done */
    };
  };

  screen("Cannot set that number data bits.\n", fromctrl);
}

/* Set number stop bits.    */
void set_stopbits(char *param)
{
  int st;
  char bf[50];

  if(
     sscanf(param, " %i", &st) == 1
    )
  {
    if(
      bios_x_init(comno, brk, parity, st, databits, baud)  /* Set */
      )
    {
      /* Setting o.k. */
      stopbits = st;                /* Save new data bits. */
      sprintf(bf, "Set %s stop bits.\n", (
					  databits == 5 &&
					  st == 2    ? "one and a half":(
					  st == 1    ? "one" :
						       "two"     )
					 )
	     );
      screen(bf, fromctrl );
      return;              /* Done */
    };
  };

  screen("Cannot set that number stop bits.\n", fromctrl);
}

  /* Use BIOS sevices to set port parameters. */
  /* Assumes presence of NewBIOS.    */
  /* Returns a flag to indicate whether it has succeded. */
int bios_x_init(int port, int brk, char par, int stop, int data, long baud)
{
  union REGS inregs, outregs;

  if( port < 1 || port > 16 )
    return FALSE;               /* Bad port specification. */
  else
    inregs.x.dx = port - 1;     /* port number. */

  inregs.h.al = (char)(brk ? 1 : 0);  /* Break parameter. */

  switch( tolower(par) )
  {
    case 'n': inregs.h.bh = 0;  /* Parity none */
	      break;
    case 'o': inregs.h.bh = 1;  /* Parity odd */
	      break;
    case 'e': inregs.h.bh = 2;  /* Parity even */
	      break;
    case 'm': inregs.h.bh = 3;  /* Parity mark */
	      break;
    case 's': inregs.h.bh = 4;  /* Parity space */
	      break;
    default:  return FALSE;     /* Bad parity specification. */
  };

  if( stop < 1 || stop > 2 )
    return FALSE;               /* Bad stop bits specification. */
  else
    inregs.h.bl = (char)(stop-1);

  if( data < 5 || data > 8 )
    return FALSE;               /* Bad databits specification. */
  else
    inregs.h.ch = (char)(data - 5);

  if(
      ( inregs.h.cl = (char)prefer_baud(&baud) ) == 255
    )
    return FALSE;               /* Bad Baud rate. */

  inregs.h.ah = 0x04;           /* Service 4, extended intialise. */
  int86(0x14, &inregs, &outregs);     /* Asynchronous services. */

  return TRUE;                  /* Done */
}

  /* Convert a (possibly abbreviated) baud rate to a full form. */
  /* Returns code of baud rate as used by NewBIOS service 04H */
  /* or -1 to indicate a value which cannot be expanded to a prefered */
  /* value. */
int prefer_baud(long *baud)
{
  int bios_param;

  if( *baud <= 32767 )       /* for short baud rate. */
    switch((int)*baud)         /* Check and extend baud rate specified */
    {
    case   11:
    case  110: *baud=110l; bios_param =0;
	       break;
    case   15:
    case  150: *baud=150l; bios_param =1;
	       break;
    case   30:
    case  300: *baud=300l; bios_param =2;
	       break;
    case   60:
    case  600: *baud=600l; bios_param =3;
	       break;
    case   12:
    case  120:
    case 1200: *baud=1200l; bios_param =4;
	       break;
    case   24:
    case  240:
    case 2400: *baud=2400l; bios_param =5;
	       break;
    case   48:
    case  480:
    case 4800: *baud=4800l; bios_param =6;
	       break;
    case   96:
    case  960:
    case 9600: *baud=9600l; bios_param =7;
	       break;
    case   19:
    case  192:
    case 1920:
    case 19200: *baud=19200l; bios_param =8;
	       break;
    case   38:
    case  384:
    case 3840: *baud=38400l; bios_param =9;
	       break;
    case   56:            /* 56 kBaud is prefered value. */
    case  560:            /* 57k6 Baud is nearest PC-style serial ports */
    case 5600:            /* are able to get.                           */
    case   57:
    case  576:
    case 5760: *baud=57600l; bios_param =10;
	       break;
    case  115:
    case 1152:
    case 11520: *baud=115200l; bios_param =11;
	       break;
    default: bios_param = -1;     /* No valid baud rate. */
	       break;
    }
    else
    {
      if( *baud == 38400l )
				     bios_param = 9;
      else if( *baud == 56000l ||
	       *baud == 57600l    ){
				     *baud = 57600l;
				     bios_param = 10;   }
      else if( *baud == 115200l )
				     bios_param = 11;
      else
				     bios_param = -1;
				     /* No valid baud rate */
    };

    return bios_param;

}

  /* Write a file to command processor. */
  /* Note that a command file can initialte another command file. */
  /* A recursive execution of a file nests up to the number of */
  /* open files allowwed.                                      */
void exec_file(char *fname)
{
  int scanned;
  char bf[120];
  FILE *file;

  if(
     sscanf(fname, " %s", bf) == 1
    )
    if( ( file = fopen(bf, "rt" ) )
	 != NULL)                          /* If find specified file. */
    {
      sprintf(bf,"Executing %s\n",fname);
      screen(bf, fromctrl);
      do
      {
	/* Read a line from file. */

	scanned = fscanf(file, "%119[^\n]\n", bf);

	/* Process it */
	if( scanned > 0 )     /* If any read in. */
	  cmdout(bf);
      }
      while( scanned != EOF );
      fclose(file);
    }
    else
    {
      sprintf(bf, "Cannot find file %s\n", fname);
      screen(bf, fromctrl);
    }
  else
  {
    sprintf(bf, "No file to execute\n");
    screen(bf, fromctrl);
  };
}

  /* Execute DOS command. */
void exec_dos(char *cmd)
{
  char buff[150];

  sprintf(buff, "Executing %s\n", cmd);
  screen(buff, fromctrl);
  if(
     system(cmd) == -1
    )
    screen("Failed to execute command.\n", fromctrl);
}

  /* Read input from keyboard */
  /* Character is written to chr. Returns TRUE if character read, */
  /* returns false if none read.      */
  /* If kb_retry is true wil wait for a charater to be read. */
  /* The two DOS functions used check for characters in the keyboard */
  /* buffer and read a character out of the buffer respectively. */
int kbdinp(char *chr, int kb_retry)
{
  if( ( bdos(0xB,0,0) & 0xFF) ||   /* If charaters available. */
	kb_retry                 ) /* Or if must wait for them */
  {
    *chr = (char)(bdos(0x7,0,0) & 0xFF); /* Get (and remember) */
						/* character. */
    if( *chr == '\r' )                  /* Dos newline character. */
      *chr = '\n';                      /* C's prefered newline. */

    return TRUE;
  }
  else
    return FALSE;
}

  /* Send data to screen(s). */
  /* scr is a flag to indicate the logical screen to use. */
void screen(char *output, enum SCREEN scr)
{
  printf("%s", output);
}

  /* Set up screen. */
void init_screen(void)
{
  /* Nothing to do here. */

}

  /* Initialise translation tables. */
  /* format is a sequence of characters representing the translations for */
  /* successive bytes. An array of 257 (1 for each character plus and end)*/
  /* pointers divides up the array.					  */
void init_xlat(void)
{
  int i;
  static char ix[XLT_SIZE];
  static char ox[XLT_SIZE];

  inxlt[0]  = ix;  /* Allocate space for */
  outxlt[0] = ox;  /* translation tables. */

/*  inxlt[0]  = (char*) malloc(XLT_SIZE);*/  /* Allocate space for */
/*  outxlt[0] = (char*) malloc(XLT_SIZE);*/  /* translation tables. */

  for(i = 0; i < 256; i++)
  {
    inxlt[i+1] = inxlt[i] + 1;	  /* Initially each pointer references */
    outxlt[i+1] = outxlt[i] + 1;  /* successive locations in xlt space */

    *inxlt[i] = (char)i;	/* Each character translates to itself. */
    *outxlt[i] = (char)i;
  };
}

  /* Look for a terminal definition file, if found execute it. */
  /* Is one specified on command line? Otherwise look for */
  /* def_file                                            */
void exec_defaults(char **argv)
{
  int handle;

  if( *(argv+1) )    /* Is defaults file specified on command line? */
  {
    if(
	( handle = open( *(argv+1), O_RDONLY | O_TEXT) ) != -1
      )
    {  /* Have been able to open file. */
      close( handle );
      exec_file( *(argv+1) );        /* Do it. */
    }
    else
    {
      screen("Cannot open file ", fromctrl);
      screen( *(argv+1), fromctrl );
      screen("\n", fromctrl );
    }
  }
  else
  {
    /* Otherwise (try to) use default default file */
    exec_file(def_file );
  }
}

  /* Open com port. */
void open_com(void)
{
  char bf[80];          /* Temporary text buffer. */

  com_inp = fopen(comname,"rb");   /* Open port */
  if( com_inp == NULL )
  {
    sprintf(bf, "Failed to open %s\nTry different port parameters.\n"
	    ,comname);
    screen(bf, fromctrl);
    return;
  };
  uncook(com_inp);  /* Set to raw data mode. */
  rewind(com_inp);

  com_out = fopen(comname,"wb");   /* Open port output */
  if( com_out == NULL )
  {
    sprintf(bf, "Failed to open %s\nTry different port parameters.\n"
	    ,comname);
    screen(bf, fromctrl);
    fclose(com_inp);
    return;
  };
  uncook( com_out);
  rewind(com_out);
}

  /* Alter mode of file from cooked (DOS default) to raw. */
  /* Prevents translation of CR/LF, trapping of CTRL-Z and */
  /* Allows read/write operations to be performed more efficiently. */
void uncook(FILE *file)
{
  union REGS inregs, outregs;

    /* Change IOCTL from Cooked to RAW data function 0x44. */
    inregs.h.ah = 0x44;
    inregs.h.al = 0x00; /* get device data */
    inregs.x.bx = fileno(file);
    intdos( &inregs, &outregs );
    inregs.x.dx=outregs.x.dx;
    inregs.h.dl=inregs.h.dl | 0x20; /* set raw data mode */
    inregs.h.dh=0;
    inregs.h.al = 0x01; /* set device data */
    intdos( &inregs, &outregs );
}

  /* Set string to upper case. */
void str_toupper(char *str)
{
  if( str-- != NULL )
    while( *++str )
      *str = toupper( *str );
}
