C *********************************************************************
C *								      *
C *	FORterm -A simple terminal emulation program for use with     *
C *	=======  MS-DOS and Serial Solutions serial port device       *
C *		 device driver NewCOM.				      *
C *								      *
C *								      *
C *********************************************************************
C *								      *
C *	      Written by  Richard Davis, 22/6/1990		      *
C *								      *
C *		    Copyright (c) 1990				      *
C *								      *
C *								      *
C *********************************************************************

C Uncomment following line to inc;lude extra debuging code
C     $DEBUG
$DECLARE

C #######################################################
C
C     Interface to commands module (FT2)
C

      interface to subroutine cmdout (cmd)
      character *256 cmd
      end

      interface to subroutine exec_file(name)
      character*256 name
      end

      interface to subroutine setup
      end

C #######################################################
C
C     Interface to support functions module (FT4)
C

      interface to function kbdinp (c, kbretry)
      logical kbdinp
      character c
      logical kbretry
      end

      interface to subroutine screen(str, source)
      character *(*) str
      integer source
      end

      interface to subroutine initxlt
      end

C #######################################################

      program FORterm

C Port parameters definition common data
      integer*1 comno,baud,parity,databits,stopbits,brk
      common /port/ comno,baud,parity,databits,stopbits,brk
C Port name
      character*8 comname
      common /port/ comname
C terminal data common
      logical echol,echos
      common /term/ echol,echos
C Internal variables common
      integer nextunit
      common /intvar/ nextunit

C Set up default values for translations
      call initxlt
C Set up other program variables
      echol = .FALSE.
      echos = .FALSE.
C Internal variable, used by exec_file
      nextunit = 7

C COM1, 1200 Baud, even parity, 7 data bits and one stop bit.
C Parameters are in form used by BIOS extended initialise.
      comname  = 'COM1    '
      comno    = 0
      baud     = 4
      parity   = 2
      databits = 2
      stopbits = 0
      brk      = 0

      call screen('\n'C,0)

      call exec_defaults

      call open_com

C Main program- is infinite loop
      do while (.TRUE.)

	call switchout

	call readin

      end do

      end

C *********************************************************************
C *								      *
C *  Open_com- opens com port					      *
C *								      *
C *********************************************************************
      subroutine open_com

C Port parameters definition common data
      integer*1 comno,baud,parity,databits,stopbits,brk
      common /port/ comno,baud,parity,databits,stopbits,brk
C Port name
      character*8 comname
      common /port/ comname

      integer io

C Binary sequential file.
      open (1, file=comname, form='binary',  iostat=io)

      if( io .NE. 0 )then
	call screen('Failed to open '//comname//'\n'C,2)
	call screen('Try different port parameters\n'C,2)
      else
	call screen('Opened serial port, called '//comname//'\n'C,2)
      end if

      end

C *********************************************************************
C *								      *
C *   Switchout- reads characters from keyboard.		      *
C *   Process any commands and send all other bytes on to port.       *
C *								      *
C *********************************************************************
      subroutine switchout
      logical kbdinp
      character*2 c

C terminal data common
      logical echol,echos
      common /term/ echol,echos

      if( kbdinp(c,.FALSE.) )then
	if ( c(1:1) .EQ. char(0) ) then
C	  nul ==> command key (eg ALT-C, F1)
	  call menuout
	else
C	  Normal character-check echo and send on
	  if( echos ) then
	    c(2:2) = ''C
	    call screen(c, 1)
	  end if
	  call outxlat(c)
	end if
      end if

      end

C *********************************************************************
C *								      *
C *  Outxlat translate a character and output it.		      *
C *								      *
C *********************************************************************
      subroutine outxlat (data)
      character data
      integer start, end, ch
C Translations common data
      character *1024 inxlt, outxlt
      integer inxltpt(0:256), outxltpt(0:256)
      common /translates/ inxlt, outxlt, inxltpt, outxltpt
      ch = ichar(data(1:1))
      start = outxltpt(ch)
      end = outxltpt(ch + 1)
      if( end .GT. start ) then
	call outstr(outxlt(start:end), end - start)
      end if
      end

C *********************************************************************
C *								      *
C *  Outstr sends count characters to the serial port. Any write      *
C *  errors are dealt with.					      *
C *								      *
C *********************************************************************
      subroutine outstr (data, count)
      character *(*) data
      integer count

      rewind(1,err=1000)
      write (1,err=1000) data(1:count)
      rewind(1,err=1000)

      return

 1000 continue
      call screen('Failed to output characters to port.\n'C,2)
      end

C *********************************************************************
C *								      *
C *  Readin- read from port and send to screen			      *
C *								      *
C *********************************************************************
      subroutine readin
      character*256 c

      call inxlat( c )
      if( c(1:1) .NE. char(0) )then
	call screen(c,0)
      end if

      end

C *********************************************************************
C *								      *
C *   Inxlat attempts to read a character from the serial port.       *
C *   If one is read in translates it. Returned string is nul-	      *
C *   terminated.								 *
C *								      *
C *********************************************************************
      subroutine inxlat (buffer)
      character *(*) buffer
      integer ch,start,end,pti,pto
      logical inpstr
      character*256 c
C Translations common data
      character *1024 inxlt, outxlt
      integer inxltpt(0:256), outxltpt(0:256)
      common /translates/ inxlt, outxlt, inxltpt, outxltpt
C terminal data common
      logical echol,echos
      common /term/ echol,echos

      pti = 1
      pto = 1

C Loop repeatedly reads from serial port until there are no more characters-
C this is because READ operations fill an internal buffer which must be
C emptied before a write operation.
       do while( inpstr(c(pti:pti)) )
C A character read in
	ch = ichar(c(pti:pti))
	pti = pti + 1
	start = inxltpt(ch)
	end = inxltpt(ch+1)-1
	if(end .GE. start)then
	  buffer(pto:pto+end-start) = inxlt(start:end)
	  pto = pto + end - start + 1
	end if
      end do
      buffer(pto:pto) = char(0)

      if( (pti .GT. 1) .AND. echol )then
C Echo to line set, send character back to port.
	call outstr(c,pti-1)
      end if

      end

C *********************************************************************
C *								      *
C *  InpStr reads a character from the serial port, and writes it     *
C *  to 'c'. The return value indicates whether any characters were   *
C *  read. If echo to line is set then send back out to port.	      *
C *								      *
C *********************************************************************
      logical function inpstr (c)
      character c
      integer io

      read (1,iostat=io) c
      inpstr = ( io .EQ. 0 )
C Rewind file (which flushes internal buffer) only when no more characters
	if(.NOT. inpstr) rewind (1)
      end

C *********************************************************************
C *								      *
C *  Menuout- reads second byte of an extended character code and     *
C *  decodes it 						      *
C *								      *
C *********************************************************************
      subroutine menuout
      logical kbdinp

      character c*256
      integer i

C     read second byte of extended code
      if( kbdinp(c,.FALSE.) ) then
	select case( ichar(c(1:1)) )
C	  ALT-C reads in a command
	  case( 46 )
	    call screen ('Enter command:'C,2)
	    read (*,'(A255)') c
	    call screen('\n'C,2)
            i = len_trim(c) + 1
            c(i:i) = char(0)
	    call cmdout(c)
C	  F1: Help option
	  case( 59 )
	    c = 'h'C
	    call cmdout(c)
C	  F2: Setup option
	  case( 60 )
            call setup
C	  F10: Quit option
	  case( 68 )
	    c = 'q'C
	    call cmdout(c)
C	  Unrecognised extended code is ignored
	  case default
	end select
      end if
      end

C *********************************************************************
C *								      *
C *   Exec_defaults-find a terminal definition file and execute it.   *
C *								      *
C *********************************************************************
      subroutine exec_defaults

      logical fexist
      integer*2 argstat
      character*80 arg

      call getarg(1,arg,argstat)

      if(argstat .GE. 1)then
	arg = arg//char(0)
	inquire(file=arg,exist=fexist)
	if(fexist)then
	  call exec_file(arg)
	  return
	else
	  call screen('Cannot find '//arg,2)
	  call screen('\n'C,2)
	end if
      end if

      inquire(file='termdef.txt',exist=fexist)
      if(fexist)then
	call exec_file('termdef.txt'C)
      else
	call screen('Cannot find termdef.txt\n'c,2)
      end if

      end


C *********************************************************************
C *                                                                   *
C *   global data for names of parameters                             *
C *                                                                   *
C *                                                                   *
C *                                                                   *
C *                                                                   *
C *                                                                   *
C *********************************************************************
      block data

      character*5 parys(0:4)
      common /parysc/ parys

      integer bauds(0:11)
      common /baudsc/ bauds

      data bauds /110,150,300,600,1200,2400,4800,
     +            9600,19200,38400,57600,115200/

      data parys /'none ','odd  ','even ',
     +            'mark ','space'/


      end
