C *********************************************************************
C *								      *
C *	FORterm -A simple terminal emulation program for use with     *
C *	=======  MS-DOS and Serial Solutions serial port device       *
C *		 device driver NewCOM.				      *
C *								      *
C *								      *
C *********************************************************************
C *								      *
C *	      Written by  Richard Davis, 22/6/1990		      *
C *								      *
C *		    Copyright (c) 1990				      *
C *								      *
C *********************************************************************
C *								      *
C *   Support functions module					      *
C *								      *
C *********************************************************************

C     $DEBUG
$DECLARE

C Interface to assembly language routine which calls interrupts
      interface to subroutine intrpt[FAR,ALIAS:"intrpt"]
     +            ( intno, regs)

      structure /regtype/
	union
	  map
	    integer*1 al,ah,bl,bh,cl,ch,dl,dh
	  end map
	  map
	    integer*2 ax,bx,cx,dx,si,di,cflag
	  end map
	end union
      end structure

      integer intno [far,reference]
      record /regtype/ regs [far,reference]
      end

C Interface to library routine which calls DOS services
      interface to function intdos (inregs, outregs )
      integer*2 intdos [C,FAR,ALIAS:"_intdos"]
      integer*4 inregs [VALUE]
      integer*4 outregs [VALUE]
      end

C *********************************************************************
C *                                                                   *
C *  kbdinp - read input from the keyboard. If kbretry is true        *
C *  then wait for a character to be typed. Otherwise return          *
C *  immediately. Return value .TRUE. if a character.                 *
C *                                                                   *
C *                                                                   *
C *********************************************************************
      function kbdinp (c, kbretry)
      logical kbdinp
      character c
      logical kbretry

      structure /regtype/
	union
	  map
	    integer*1 al,ah,bl,bh,cl,ch,dl,dh
	  end map
	  map
	    integer*2 ax,bx,cx,dx,si,di,cflag
	  end map
	end union
      end structure

      record /regtype/ registers
      integer*2 dummy
      integer*2 intdos

      registers.ah = 16#0B


      dummy = intdos( locfar(registers), locfar(registers) )
      if( (registers.al .NE. 0) .OR. kbretry )then
C       get a character from keyboard
        registers.ah = 16#07
        dummy = intdos( locfar(registers), locfar(registers) )
        c = char( registers.al )
C       Convert keyboard end of line (CR) to internal end of line (LF)
        if( c .EQ. char(13) )
     +    c = char(10)
        kbdinp = .TRUE.
      else
        kbdinp = .FALSE.
      end if

      end

C *********************************************************************
C *								      *
C *   Initxlt initialises translation tables. Inxlt/outxlt contain    *
C *   translations, inxltpt and outxltpt contain pointers into	      *
C *   these translations.					      *
C *								      *
C *********************************************************************
      subroutine initxlt
C Translations common data
      character *1024 inxlt, outxlt
      integer inxltpt(0:256), outxltpt(0:256)
      common /translates/ inxlt, outxlt, inxltpt, outxltpt

      integer i

      do i = 0, 255
	inxlt(i+1:i+1) = char(i)
	outxlt(i+1:i+1) = char(i)
	inxltpt(i) = i+1
	outxltpt(i) = i+1
      end do

      inxltpt(256) = 257
      outxltpt(256) = 257

      end


C *********************************************************************
C *								      *
C *  Screen sends the characters of its first operand to the console  *
C *  (unit *). The string is nul-terminated. The second parameter     *
C *  is a destination code.					      *
C *								      *
C *********************************************************************
      subroutine screen (str, source)
      character *(*) str
      integer source
      integer i

      do i = 1, 32767
	if ( str(i:i) .EQ. char(0) ) return
	if ( str(i:i) .EQ. char(10) ) then
	  write (*,'(/,'' '',\)')
	else
	  write (*,'(A1,\)') str(i:i)
	end if
      end do
      end

C *********************************************************************
C *								      *
C *  BIOS_x_init, checks its parameters, and if they fall within      *
C *  the correct ranges uses them to set the serial port via	      *
C *  the BIOS services. Returns .TRUE. if works, else .FALSE.	      *
C *								      *
C *********************************************************************
      logical function BIOS_x_init(port, brk, par, stop, data, Baud)
      integer*1 port, brk, par, stop, data, Baud

      structure /regtype/
	union
	  map
	    integer*1 al,ah,bl,bh,cl,ch,dl,dh
	  end map
	  map
	    integer*2 ax,bx,cx,dx,si,di,cflag
	  end map
	end union
      end structure

      record /regtype/ registers

C Return false if any parameters bad
      BIOS_x_init = .FALSE.

      registers.ah = 4
      if( (port .LT. 0) .OR. (port .GT. 15) ) return
      registers.dx = port
      if( (brk .LT. 0) .OR. (brk .GT. 1) ) return
      registers.al = brk
      if( (par .LT. 0) .OR. (par .GT. 4) ) return
      registers.bh = par
      if( (data .LT. 0) .OR. (data .GT. 3) ) return
      registers.ch = data
      if( (stop .LT. 0) .OR. (stop .GT. 1) ) return
      registers.bl = stop
      if( (Baud .LT. 0) .OR. (Baud .GT. 11) ) return
      registers.cl = Baud

      call intrpt( 16#14, registers )

C Return true for good numbers
      BIOS_x_init = .TRUE.

      end

C *********************************************************************
C *								      *
C *  tolower- returns its operand, converted to lower case if	      *
C *  it is an upper case character.				      *
C *								      *
C *********************************************************************
      character function tolower (c)
      character c

      if( (c .GE. 'A') .AND. (c .LE. 'Z') ) then
	tolower = char(ichar(c) + 32)
      else
	tolower = c
      end if
      end

C *********************************************************************
C *								      *
C *   Parse_str_def reads through a command string, and builds a      *
C *   a sequence of characters from it. String is of form:	      *
C *	<strdef2>={|<chardef>|<quotedstring2>}[,<strdef2>]	      *
C *	<chardef>=decimal 0..255				      *
C *	<quotedstring2>={''|""|<quotedstring1>} 		      *
C * <quotedstring1>={'1 or more characters'|"1 or more characters"}   *
C *   The function returns true if string parses ok. Operand is nul   *
C *   terminated, and sequence is terminated by end of string or :    *
C *								      *
C *********************************************************************
      logical function parse_str_def (str, opseq ,num)
      integer num
      character *256 str, opseq
      character thisquote
      logical is_quote
      logical is_digit
      integer i,ci

      call dropspace(str)
      parse_str_def = .TRUE.
      num = 1

      do while( (str(1:1) .NE. char(0)) .AND.
     +		(str(1:1) .NE. ':'    )       )
	parse_str_def = .FALSE.
	call dropspace(str)

	if( is_quote(str) )then
C	Read characters between quotes into opseq
	  thisquote = str(1:1)
	  str = str(2:)
	  do while( ( str(1:1) .NE. char(0)  ) .AND.
     +		    ( str(1:1) .NE. thisquote) )
	    opseq(num:num) = str(1:1)
	    num = num + 1
	    str = str(2:)
	  end do
C	  Check presence of final quote mark
	  if( str(1:1) .EQ. char(0) ) return
	  str = str(2:)
	else if(is_digit(str))then
C	  Try and interpret characters as a number
	  i=1
	  do while( is_digit( str(i:i) ) )
	    i = i+1
	  end do
	  read (str(1:i-1),'(I3)') ci
	  if( ci .GT. 255 ) return
	  opseq(num:num) = char(ci)
	  num = num + 1
	  str = str(i:)
	end if
	call dropspace (str)
	if( str(1:1) .NE. char(0) )then
	  if( str(1:1) .EQ. ',' ) then
	    str = str(2:)
	  else
	    if( str(1:1) .NE. ':' )return
	  end if
	end if
	parse_str_def = .TRUE.
      end do
      num = num - 1
      end

C *********************************************************************
C *								      *
C *   Is_digit- returns true if operand is '0','1', etc		      *
C *								      *
C *********************************************************************
      logical function is_digit (str)
      character str

      is_digit = ( scan(str,'0123456789') .NE. 0 )
      end

C *********************************************************************
C *								      *
C *  Is_quote returns true if operand is ' or "                       *
C *								      *
C *********************************************************************
      logical function is_quote (str)
      character str
      is_quote = ( (str .EQ. '''') .OR.
     +		   (str .EQ. '"' )	)
      end

C *********************************************************************
C *								      *
C *  Dropspace removes and blanks from begining of operand string     *
C *								      *
C *********************************************************************
      subroutine dropspace (str)
      character *256 str

      do while( str(1:1) .EQ. ' ' )
	str = str(2:)
      end do

      end

C *********************************************************************
C *								      *
C *  Screscstr- unpacks a string of characters into characters and    *
C *  and numbers in order to print them.			      *
C *								      *
C *********************************************************************
      subroutine screscstr (str, num, source)
      character*256 str
      integer num, source
      integer i
      logical first
      logical wasprintable
      character*10 tmps

      first = .TRUE.
      wasprintable = .FALSE.

      do i = 1, num
	if( (str(i:i) .LT. char(32) ) .OR.
     +	    (str(i:i) .GT. char(127)) .OR.
     +	    (str(i:i) .EQ. '"'	    )	   )then
C	Deal with non-printable character
	  if(wasprintable) call screen('"'C,source)
	  wasprintable = .FALSE.

	  if(first)then
	    first = .FALSE.
	  else
	    call screen(','C,source)
	  end if

	  write(tmps,'(I3,A1)') ichar(str(i:i)),char(0)
	  call screen(tmps,source)

	else
C	Deal with printable character
	  if(wasprintable)then
	  else
	    if( first ) then
	      first = .FALSE.
	    else
	      call screen(','C,source)
	    end if
	    call screen('"'C,source)
	  end if

	  call screen(str(i:i)//char(0),source)
	  wasprintable = .TRUE.
	end if
      end do
      if( (.NOT. first) .AND. wasprintable)
     +	 call screen ('"'C,source)

      end
