PROGRAM Pasterm;
{ ***********************************************************************
  *									*
  *	Pasterm- A simple terminal emulation program to demonstrate	*
  *	=======  the use Serial Solutions COM port device driver under	*
  *		 MS-DOS 						*
  *									*
  ***********************************************************************
  *									*
  *	Written by Richard Davis, 8/6/1990, in TurboPascal Version 4.0	*
  *			  Copyright (c) 1990				*
  *									*
  *********************************************************************** }


{$M 8192,0,16384}
{$R+}
USES
  Dos, Crt;

CONST
  Def_File = 'termdef.txt';     {Default definition file}
  Xlt_Size = 1024;		{Max number bytes in translate table}
  Newl	:Char = #10;		{Newline Character}

  ParNames: Array[0..4] Of String[5] = ('none','odd','even','mark','space');

  Bauds: Array[0..11] Of LongInt = ( 110,   150,  300,  600,   1200,
                                     2400,  4800, 9600, 19200, 38400,
                                     57600, 115200                   );

TYPE
  Screen_type = (Fromport, Fromecho, Fromctrl);
			    {Defines source of any screen o/p}
  Echot  = (Echos, Echol);  {Echo strategies; to screen, to line}

  Xlt_Table = Array[0..Xlt_Size] Of Char;     {Table of translations}
  Xlt = RECORD
	  Table :^Xlt_Table;
	  Point :Array[0..256] Of Integer;	  {Pointers into above}
	END;
VAR
  ComNo 	:Integer;	{n-1 for COMn, -1 for not a com port}
  ComName	:String[8];	{Device name}

  Baud		:Byte;		{Port parameters. }
  Parity	:Byte;		{Format is as operands of a call to }
  DataBits	:Byte;		{the BIOS extended initialise service}
  StopBits	:Byte;
  Brk		:Byte;

  Echo		:SET OF Echot;	{Echo type in operation}

  Com_Inp,			        {Input and  }
  Com_Out	:File Of Char;		{output serial port files }

  In_Xlt,
  Out_Xlt	:Xlt;		{Input & output translations }

  { Screen Output }
PROCEDURE Screen (Oput :String; Scr :Screen_type);
VAR
  i	:Integer;
BEGIN
  FOR i := 1 to Length( Oput ) DO
    IF
      Oput[i] = Newl		{Newl means newline}
    THEN
      Writeln
    ELSE
      Write(Oput[i]);
END;

  {Write to translation tables a replacement for a key}
  {Returns true is successful}
PROCEDURE Set_Xlt_Low (VAR Xtable: Xlt; Key: Char; Rep: String);
VAR
  Dlength, Keyn, Ptr: Integer;
BEGIN
  WITH Xtable DO BEGIN
    Keyn := Ord(Key);
    Dlength := Length(Rep) -                    {Change in length of entry}
               ( Point[Keyn+1] - Point[Keyn] );

    IF    {Check enough space}
      ( Point[256] - Point[0] + Dlength ) > Xlt_Size
    THEN BEGIN
      Screen('Not enough space for that translation, sorry.'+Newl,
             Fromctrl);
      Exit;
      END;

    IF
      Dlength <> 0     {If must change of table}
    THEN BEGIN
      IF
        Dlength > 0      {If must stretch of table}
      THEN
        FOR Ptr := Point[256] DOWNTO Point[Keyn+1] DO
          Table^[Ptr+Dlength] := Table^[Ptr]
      ELSE                {If must contract table}
        FOR Ptr := Point[Keyn+1] TO Point[256] DO
          Table^[Ptr+Dlength] := Table^[Ptr];

      FOR Ptr := Keyn + 1 TO 256 DO          {ALter pointer}
        Point[Ptr] := Point[Ptr] + Dlength;
      END;

    {Write in new representation}
    FOR Ptr := 0 TO Length( Rep ) - 1 DO
      Table^[Point[Keyn]+Ptr] := Rep[Ptr + 1];
  END;
END;

  {Procedure takes first character away from a string}
PROCEDURE Tail (VAR s: String);
BEGIN
{$IFDEF TAILMON}
  IF Length(s) > 0 THEN
    Write('<',s[1]);
{$ENDIF}
  s := Copy(s,2,255);
END;

  {Remove any spaces or tabs at start of a string}
PROCEDURE DropSpace(VAR s: String);
BEGIN
  WHILE
    (Length(s) > 0) AND
    ( s[1] In [' ',#9] )
  DO
    Tail(s);  {Strip off characters}
END;

  {Returns true if string is Not empty and first character is ' or " }
FUNCTION Is_Quote (S: String):Boolean;
BEGIN
  IF
    Length( S ) > 0
  THEN
    Is_Quote := S[1] In [#34,#39]
  ELSE
    Is_Quote := false;
END;

  {Returns true if first character of operand string is a digit}
  {and string is of non-zero length}
FUNCTION Is_Digit (VAR s: String):Boolean;
BEGIN
  Is_Digit := ( Length( s ) > 0 ) AND
              ( s[1] >= '0' ) AND
              ( s[1] <= '9' )
END;

  {Read through a command string parameter and build a sequence}
  {of characters from it. String is of form <String_def2>, above}
  {Termination is ':' or end of string}
  {If termination is ':' this is returned as first character of Str}
  {Returns flag true for sucessful decode}
FUNCTION Parse_Str_Def (VAR Str, Opseq: String):Boolean;
VAR
  ThisQuote: Char;
  Num,Code:  Integer;
  DigitStr:  String;
BEGIN
  Opseq := '';
  DropSpace(Str);
  Parse_Str_def := true;
  WHILE
    ( Length(Str) > 0 ) AND
    ( Str[1] <> ':' )         {Terminator for sequences}
  DO BEGIN
    Parse_Str_Def := false;
    DropSpace(Str);

    IF
      Is_Quote(Str)
    THEN BEGIN
      ThisQuote := Str[1];
      Tail(Str);
      WHILE
        ( Length( Str ) > 0 ) AND
        ( Str[1] <> ThisQuote )
      DO BEGIN
        Opseq := Opseq + Str[1];
        Tail(Str);
      END;
      IF
        Length( Str ) = 0  {Bad return, no closing quote}
      THEN
        Exit;
      Tail(Str);  {Drop Closing Quote}
    END
    ELSE IF
      Is_Digit(Str)
    THEN BEGIN
      DigitStr := '';
      REPEAT
        DigitStr := DigitStr + Str[1];
        Tail(Str);
      UNTIL
        NOT IS_Digit(Str);
      {$R-}
      Val(DigitStr, Num, Code);
      {$R+}
      IF
        ( Code <> 0 ) OR
        ( Num > 255 ) OR
        ( Num < 0   )
      THEN
        Exit;              {Bad Number Specification}
      Opseq := Opseq + Chr(Num);
    END;
    DropSpace(Str);
    IF
      Length( Str ) > 0
    THEN
      IF
        Str[1] = ','
      THEN
        TAil( Str )      {Strip off comma}
      ELSE
        IF
          Str[1] = ':'
        THEN               {Do nothing}
        ELSE
          Exit;            {Bad separator}

    Parse_Str_Def := true;    {Good parameter}
  END;
END;

  {Read a character from keyboard. Returns true if a character read}
  {returns false if none read. If Kb_Retry is true the function}
  {waits for user to press a key, otherwise routine returns immediately}
FUNCTION KbdInp (VAR c :Char; Kb_Retry :Boolean): Boolean;
BEGIN
  IF
    Kb_Retry OR KeyPressed
  THEN BEGIN
    c := ReadKey;
    IF c = #13 THEN c := newl;  {Our internal newline character}
    KbdInp := true;
  END
  ELSE
    KbdInp := false;
END;

  {Send a string to serial port}
  {Must deal here with any errors}
PROCEDURE OutStr (s :String);
VAR
  i: Integer;
BEGIN
{$IFDEF OUTMONIT}
  Write('OutStr ',Length(s),' bytes:');
  FOR i := 1 TO Length(s) DO
    IF ( s[i] <= #32 ) OR
       ( s[i] >= #127 )
    THEN Write(',#',Ord(s[i]),',')
    ELSE Write(s[i]);
  Writeln;
{$ENDIF}
{$I-}
  FOR i := 1 TO Length( s ) DO
    Write(Com_Out, s[i]);
{$I+}
  IF
    IoResult <> 0	{Write failure}
  THEN
    Screen('Failed to output characters to port.'+Newl,Fromctrl);
END;

  {Close com files. Checks that they are open first}
PROCEDURE Close_Com;
BEGIN
  {$I-}                                 {Disable I/O checking}
  IF
    FileRec(Com_Inp).Mode = fmInput
  THEN
    Close(Com_Inp);
  IF
    FileRec(Com_Out).Mode = fmOutput
  THEN
    Close(Com_Out);
  {$I+}
END;

  {Open COM port for input, output}
PROCEDURE Open_Com;

    { Change the DOS mode of a text file from cooked to raw }
  PROCEDURE Uncook (VAR Filename: FileRec);
  VAR
    ioregs :Registers;
  BEGIN
    WITH ioregs, Filename DO BEGIN
      ax := $4400;			{ Function 44H, subservice 0H }
      bx := Handle;			{ get device data.	      }
      MsDos( ioregs );

      dh := 0;				{ Set high byte of data zero }
      dl := dl OR $20;			{ Set bit 5 of low byte-raw mode}

      ax := $4401;			{ Subservice 1H, Set device data.}
      bx := Handle;
      MsDos( ioregs );
    END;
  END;
VAR
  IO1,IO2	:Boolean;
BEGIN
  {$I-}
  Assign(Com_Inp,ComName);	{Input Char file..}
  Reset(Com_inp);
  IO1 := IoResult <> 0;          {Reset failed}

  Assign(Com_Out,ComName);	{Output Char file..}
  ReWrite(Com_out);
  IO2 := IoResult <> 0;          {ReWrite failed}
  {$I+}
  IF
    IO1 OR IO2
  THEN BEGIN	{ Opens have failed }
    Screen('Failed to open '+ComName+Newl,FromCtrl);
    Screen('Try different port parameters'+Newl,Fromctrl);
    Close_Com;      {Close files}
  END
  ELSE BEGIN	{ Opens ok }
    UnCook(FileRec(Com_Inp));		  {Set RAW data mode}
    UnCook(FileRec(Com_Out));		  {Set RAW data mode}
  END;
  {$IFDEF OPENMON}
  Write('Open: Com_Inp Handle=',FileRec(Com_Inp).Handle);
  Write(', Mode=',FileRec(Com_Inp).Mode);
  Write(', Com_Out Handle=',FileRec(Com_Out).Handle);
  Writeln(', Mode=',FileRec(Com_Out).Mode);
  {$ENDIF}
END;

  {Use BIOS services to set port parameters}
  {Assumes presence of NewBIOS for extended initialise}
  {Returns a flag to indicate whether it has succeded}
FUNCTION BIOS_X_Init (Port: Integer; Brk, Par, Stop, Data, Baud: Byte)
                      :Boolean;
VAR
  ioregs :Registers;
BEGIN
  WITH ioregs DO BEGIN
    BIOS_X_Init := false;
    {Check port number}
    IF
      ( Port < 0  ) OR
      ( Port > 15 )
    THEN
      Exit;             {Bad Port number}
    DX := Port;
    {Check break setting}
    IF
      ( Brk < 0  ) OR
      ( Brk > 1 )
    THEN
      Exit;             {Bad Brk setting}
    AL := Brk;
    {Check parity}
    IF
      ( Par < 0  ) OR
      ( Par > 4 )
    THEN
      Exit;             {Bad parity number}
    BH := Par;
    {Check stop bits}
    IF
      ( Stop < 0  ) OR
      ( Stop > 1 )
    THEN
      Exit;             {Bad Stop number}
    BL := Stop;
    {Check word length}
    IF
      ( Data < 0  ) OR
      ( Data > 3 )
    THEN
      Exit;             {Bad Data number}
    CH := Data;
    {Check Baud rate}
    IF
      ( Baud < 0  ) OR
      ( Baud > 11 )
    THEN
      Exit;             {Bad Baud number}
    CL := Baud;

    AH := $4;            {service 4, extended initialise}

    {$IFDEF BIOSXMON}
    Writeln('BIOS_X_Init, AX=',AX,', BX=',BX,', CX=',CX,', DX=',DX);
    {$ENDIF}

    {Do initialisation}
    Intr($14, ioregs);   {Int 14H is BIOS Asynchronous services interrupt}

    BIOS_X_Init := true;  {This service does not generate any error codes}
    END;
END;

  {Display Help information on screen}
PROCEDURE Help;
CONST
  help1  :String = 'Functions keys:';
  help2  :String =
' F1             Help                   F10            Quit';
  help3  :String =
' F2             Setup';
  help4  :String =
'Type ALT-C (ie press "C" key while "alt" key is held down) to issue command.';
  help5  :String =
'Commands are:';
  help6  :String =
' H              help                   Q              quit';
  help7  :String =
' E{n|s|l|ls}    set echo screen,line   C<portname>    set port (e.g COM2)';
  help8  :String =
' D{<Baudrate>   set Baud rate          P{n|o|e|m|s}   set parity';
  help9  :String =
' I<char>:<strn> translate input        O<char>:<strn> translate output';
  help10 :String =
' ( <char> is a decimal number or character in quotation marks.';
  help11 :String =
'   <strn> is a sequence of numbers & quoted strings separated by commas, Eg';
  help12 :String =
'   o12:12,"New Page",13,10,10                                               )';
  help13 :String =
' X<filename>    execute file           R<Command...>  Run DOS command';
  help14 :String =
' T<strn>        tranmit sequence       ';
  help15 :String = #39;
  help16 :String =
                                          '              Comment';
BEGIN
  Screen(help1+newl,Fromctrl);
  Screen(help2+newl,Fromctrl);
  Screen(help3+newl,Fromctrl);
  Screen(help4+newl,Fromctrl);
  Screen(help5+newl,Fromctrl);
  Screen(help6+newl,Fromctrl);
  Screen(help7+newl,Fromctrl);
  Screen(help8+newl,Fromctrl);
  Screen(help9+newl+newl,Fromctrl);
  Screen(help10+newl,Fromctrl);
  Screen(help11+newl,Fromctrl);
  Screen(help12+newl+newl,Fromctrl);
  Screen(help13+newl,Fromctrl);
  Screen(help14,Fromctrl);
  Screen(help15+help16+newl,Fromctrl);
END;

PROCEDURE Exec_File (Name: String); FORWARD;

  {Parse a string, output without translation}
PROCEDURE Transmit (VAR Tns: String);
VAR
  Op,Tmp:    String;
BEGIN
  Tmp := Tns;
  IF
    Parse_Str_Def(Tns, Op)  {Decode operand string}
  THEN
    IF
      Length(Tns) = 0       {Termination was end of string}
    THEN BEGIN
      OutStr(Op);             {Send to port}
      Exit;
    END;
  Screen('Could not decode:'+Tmp+newl,Fromctrl)
END;

(*  Set a translation
    Xtable is the translation record to set.
    CmdStr is a string containing the parameters to use.
    Format is as follows:
    <String_Def1>:<String_Def2>
    where <String_Def1>={<Char_Def>|<Quoted_string1>}[,<String_Def2>]
          <String_Def2>={|<Char_Def>|<Quoted_String2>}[,<String_Def2>]
          <Char_Def>=decimal number 0..255
          <Quoted_String1>='1 or more characters'
          <Quoted_String2>={|<Quoted_String1>}                   *)
PROCEDURE Set_Xlt (VAR Xtable: Xlt; VAR Params: String);
VAR
  Key:     Char;
  Rep:     String;
FUNCTION EscStr( S: String):String;
  VAR
    i:     Integer;
    Tmp:   String;
    Nums:  String[3];
  BEGIN
    Tmp := '';
    FOR i := 1 TO Length(S) DO
      IF ( s[i] <= #32 ) OR
         ( s[i] > #128 )
      THEN BEGIN
        Str(Ord(S[i]), Nums);
        Tmp := Tmp + '#' + Nums + ',';
        END
      ELSE
        Tmp := Tmp + S[i] + ',';
    EscStr := Tmp;
  END;

BEGIN
  IF
    NOT Parse_Str_Def(Params, Rep)
  THEN BEGIN
    Screen('Cannot understand key string.'+Newl,Fromctrl);
    Exit;
    END;
  IF
    Length( rep ) <> 1
  THEN BEGIN
    Screen('Key string must be a single character'+Newl,Fromctrl);
    Exit;
    END
  ELSE
    Key := Rep[1];

  IF
    ( Length( Params ) > 0 ) AND
    ( Params[1] <> ':' )
  THEN BEGIN
    Screen('Strings must be separated by '':''.'+Newl,Fromctrl);
    Exit;
    END;

  Tail( Params );

  IF
    NOT Parse_Str_Def(Params, Rep)
  THEN BEGIN
    Screen('Cannot understand replacement string.'+Newl,Fromctrl);
    Exit;
    END;

  Screen('From '+EscStr(Key)+' to '+EscStr(Rep)+Newl,Fromctrl);

  Set_Xlt_Low(Xtable, Key, Rep);
END;

  { Sets echo parameters for terminal }
  { Echo can be to screen, to line, or both }
  { These are "" or "n", "s", "l" and "ls" or "sl" }
PROCEDURE Set_Echo (VAR Ech: String);
BEGIN
  Echo := [];  {No echo}
  WHILE
    Length( Ech ) > 0 DO BEGIN
      CASE
        UpCase(Ech[1])
      OF
        'N':  { Set no echo }
             Echo := [];
        'L':  { Set line echo }
             Echo := Echo + [Echol];
        'S':  { Set Screen echo }
             Echo := Echo + [Echos];
        ' ':  { Ignore spaces }
             ;
      ELSE
        Screen('Unrecognised character in echo setting. Ignored'+Newl,
               Fromctrl);
      END;
      Tail(Ech);
    END;

  IF
    Echo = []
  THEN
    Screen('Echo off.'+Newl,Fromctrl)
  ELSE IF
    Echo = [Echol]
  THEN
    Screen('Echo to line only.'+Newl,Fromctrl)
  ELSE IF
    Echo = [Echos]
  THEN
    Screen('Echo to screen only.'+Newl,Fromctrl)
  ELSE
    Screen('Echo to screen and line.'+Newl,Fromctrl);
END;

  { Set COM port }
  { Parameter is string eg 'COM2', if of form 'COMn' n is stripped off}
  { and written to comno}
PROCEDURE Set_Com (VAR Name: String);
VAR
  i,Len:    Integer;
  Code:     Integer;
  Com:      LongInt;
BEGIN
  DropSpace( Name );
  Len := Length( Name );
  FOR i := Length( Name ) DOWNTO 1 DO BEGIN    {Convert to upper case}
    Name[i] := UpCase( Name[i] );              {Find end of string}
    IF                                         {By first space character}
      Name[i] = ' '
    THEN
      Len := i;
    END;

  IF
    Len > 8
  THEN BEGIN
    Screen('Device name to long- truncating'+Newl,Fromctrl);
    Len := 8;
    END;

  Name[0] := Chr(Len);   {Set string to length}

  IF
    Len > 0
  THEN BEGIN
    Close_Com;

    ComName := Name;

    {Look for a port number}
    IF
      Pos('COM',Name) = 1
    THEN BEGIN
      Name := Copy(Name,4,Len-3);
      {$R-}
      Val(Name, Com, Code);
      {$R+}
      IF
        ( Code <> 0 ) OR
        ( Com < 1  ) OR
        ( Com > 16 )
      THEN
        ComNo := -1        {Bad com port number}
      ELSE
        ComNo := Com-1;    {Good com port number}
      END
    ELSE
      ComNo := -1;         {Not a 'COM' port}
    Open_Com;
    END;
  Str(ComNo, Name);
  Screen('Port is '+ComName+Newl,Fromctrl);
END;

  {Set Parity}
PROCEDURE Set_Parity (VAR Param: String);
VAR
  P: Byte;
BEGIN
  DropSpace( Param );
  IF
    Length( Param ) > 0
  THEN BEGIN
    P := Pos(UpCase(Param[1]),'NOEMS');
    IF
      P <> 0      {Ie a match found}
    THEN BEGIN    {Set parity}
      P := P - 1;
      IF
        BIOS_X_Init(ComNo, Brk, P, StopBits, DataBits, Baud)
      THEN BEGIN
        Parity := P;
        Screen('Parity set to '+ParNames[Parity]+Newl,Fromctrl);
        Exit;
        END;
      END;
    Screen('Cannot set that parity'+Newl,Fromctrl);
    END;
  Screen('Parity is '+ParNames[Parity]+Newl,Fromctrl);
END;

  {Set Baud Rate}
PROCEDURE Set_Baud (VAR Param: String);
CONST
  BaudSynms: Array[1..52] Of LongInt = ( 110, 11, 0,
                                         150, 15, 0,
                                         300, 30, 0,
                                         600, 60, 0,
                                         1200, 120, 12, 0,
                                         2400, 240, 24, 0,
                                         4800, 480, 48, 0,
                                         9600, 960, 96, 0,
                                         19200, 1920, 192, 19, 0,
                                         38400, 3840, 384, 38, 0,
                                         56000, 5600, 560, 56,
                                         57600, 5760, 576, 57, 0,
                                         115200, 11520, 1152, 115,0 );

    {Convert a Baud rate (possibly abbreviated) to a Baud number}
    {-1 indicates a bad baud rate}
  FUNCTION Prefer_Baud (Baud: LongInt):Byte;
  VAR
    i:     Integer;
    Count: Byte;
  BEGIN
    Prefer_Baud := $FF;
    Count := 0;
    FOR i := 1 TO 52 DO BEGIN
      IF
        BaudSynms[i] = 0        {End of a group of synonyms}
      THEN
        Count := Count + 1;
      IF
        BaudSynms[i] = Baud     {Found a match}
      THEN
        Prefer_Baud := Count;
      END;
  END;

VAR
  B:    Byte;
  Code: Integer;
  Bn:   Longint;
BEGIN
  DropSpace( Param );
  IF
    Length( Param ) > 0
  THEN BEGIN
    {$R-}
    Val(Param, Bn, Code);
    {$R+}
    B := Prefer_Baud(Bn);
    IF
      ( Code = 0 ) AND
      BIOS_X_Init(ComNo, Brk, Parity, StopBits, DataBits, B)
    THEN BEGIN       {Good Baud string}
      Baud := B;
      Str(Bauds[Baud], Param);
      Screen('Baud rate set to '+Param+Newl,Fromctrl);
      END
    ELSE
      Screen('Cannot set that Baud rate.'+Newl,Fromctrl);
    END
  ELSE BEGIN
    Str(Bauds[Baud], Param);
    Screen('Baud rate is '+Param+Newl,Fromctrl);
    END;
END;

  { Set number data bits (word length) }
PROCEDURE Set_DataBits (VAR Param: String);
VAR
  Num:    Longint;
  Code:   Integer;
  D:      Byte;
BEGIN
  DropSpace(Param);
  IF
    Length( Param ) > 0
  THEN BEGIN
    {$R-}
    Val(Param, Num, Code);
    {$R+}
    IF
      ( Code = 0 ) AND
      ( Num >= 5 ) AND
      ( Num <= 8 )
    THEN BEGIN
      D := Num - 5;
      IF
        BIOS_X_Init(Comno, Brk, Parity, StopBits, D, Baud)
      THEN BEGIN
        DataBits := D;
        Str(DataBits+5,Param);
        Screen('Set number data bits to '+Param+Newl,Fromctrl);
        Exit;
        END;
      END;
    Screen('Cannot set that number data bits'+Newl,Fromctrl);
    END
  ELSE BEGIN
    Str(DataBits+5,Param);
    Screen('Number data bits is '+Param+Newl,Fromctrl);
    END;
END;

  { Set number stop bits }
PROCEDURE Set_StopBits (VAR Param: String);
VAR
  Num:    Longint;
  Code:   Integer;
  S:      Byte;
BEGIN
  DropSpace(Param);
  IF
    Length( Param ) > 0
  THEN BEGIN
    {$R-}
    Val(Param, Num, Code);
    {$R+}
    IF
      ( Code = 0 ) AND
      ( Num >= 1 ) AND
      ( Num <= 2 )
    THEN BEGIN
      S := Num - 1;
      IF
        BIOS_X_Init(Comno, Brk, Parity, S, DataBits, Baud)
      THEN BEGIN
        StopBits := S;
        Str(StopBits+1,Param);
        Screen('Set number stop bits to '+Param+Newl,Fromctrl);
        Exit;
        END;
      END;
    Screen('Cannot set that number stop bits'+Newl,Fromctrl);
    END
  ELSE BEGIN
    Str(StopBits+1,Param);
    Screen('Number stop bits is '+Param+Newl,Fromctrl);
    END;
END;

  { Execute command }
PROCEDURE CmdOut (Command :String);
VAR
  c:      Char;
  DCode:  String[5];
BEGIN
  IF Length(Command) > 0 THEN BEGIN
    c := Command[1];
    Tail(Command);
    CASE UpCase(c) OF
      'Q': BEGIN {Quit program}
	     Screen('Quitting program.'+Newl,Fromctrl);
  	     Close_Com;
	     Halt(0);
           END;
      'I': BEGIN {Set Input Translation}
             Screen('Setting input translation.'+Newl,Fromctrl);
             Set_Xlt(In_Xlt,Command);
	   END;
      'O': BEGIN {Set Output translation}
             Screen('Setting output translation.'+Newl,Fromctrl);
             Set_Xlt(Out_Xlt,Command);
	   END;
      'H': BEGIN {Help Option}
             Help;
	   END;
      'C': BEGIN {Set COM Port}
             Set_Com(Command);
	   END;
      'B': BEGIN {Set Baud rate}
             Set_Baud(Command);
	   END;
      'P': BEGIN {Set parity}
             Set_Parity(Command);
	   END;
      'D': BEGIN {Set Databits}
             Set_DataBits(Command);
	   END;
      'S': BEGIN {Set stop bits}
             Set_StopBits(Command);
           END;
      'X': BEGIN {eXecute file as commands}
             Exec_File(Command);
	   END;
      'R': BEGIN {Run DOS command}
             Screen('Executing DOS command-'+Command+Newl,Fromctrl);
             Exec('\COMMAND.COM','/C '+Command);
             IF
               DosError <> 0
             THEN
               Screen('Could not execute command'+newl,Fromctrl)
             ELSE BEGIN
               Str(DosExitCode, DCode);
               Screen('Return code was '+DCode+newl,Fromctrl);
             END;
	   END;
      'E': BEGIN {Set echo}
             Set_Echo( Command );
	   END;
      'T': BEGIN {Transmit sequence}
             Transmit(Command);
	   END;
      #39: BEGIN {(Single quote) Remark}
             Screen(Command+newl,Fromctrl); {Simply echo}
	   END;
    ELSE         {Could not decode}
      Screen('Command not recognised'+newl,Fromctrl);
    END;
  END;
END;

  {Open a file, read lines from it and execute them as commands}
PROCEDURE Exec_File (Name: String);
VAR
  f: Text;
  Line:  String;
BEGIN
  DropSpace(Name);
  Assign(f, Name);
{$I-}
  Reset(f);
{$I+}
  IF
    IOResult = 0
  THEN BEGIN
    Screen('Executing file '+Name+Newl,Fromctrl);
    WHILE
      NOT Eof(f)
    DO BEGIN
      Readln(f, Line);         {Get line}
      DropSpace(Line);
      CmdOut(Line);            {Execute it}
    END;
    Close(f);
  END
  ELSE
    Screen('Could not open file '+Name+newl,Fromctrl);
END;

  { Find a definition file, attempt to open and execute }
PROCEDURE Exec_Defaults;
VAR
  Tmp:  Text;
BEGIN
  IF
    ParamCount > 0   {If a file has been specified}
  THEN BEGIN
    Assign(Tmp, ParamStr(1));
    {$I-}
    Reset(Tmp);
    {$I+}
    IF
      IOResult = 0
    THEN BEGIN       {Good file-use}
      Close(Tmp);
      Exec_File(ParamStr(1));   {Execute file}
      Exit;                     {Done}
      END
    ELSE
      Screen('Could not open file '+ParamStr(1)+Newl,Fromctrl);
  END;
  Exec_File(Def_File);           {Execute file}
END;

  { Initialise translation table }
PROCEDURE Init_Xlt (VAR X: Xlt);
VAR
  i	:Integer;
BEGIN
  IF
    MaxAvail < Xlt_Size
  THEN BEGIN
    Screen('Not enough memory to initialise translation table.'+Newl,
           Fromctrl);
    CmdOut('q');            { Quit Program }
  END
  ELSE
  WITH X DO BEGIN
    New(Table); 	{Grab storage from Heap}
    FOR i := 0 TO 255 DO BEGIN
      Point[i] := i;
      Table^[i] := Chr(i);
    END;
    Point[256] := 256;	{End Pointer}
  END;
END;

  {Menu function. Operands are title and list of options.}
  {Returns user selection}
  {Initiate set of menus to set up serial port}
PROCEDURE SetUp;
TYPE
  OptRec = Array[0..8] Of String[80];
CONST
  PortMn:  OptRec = ('COM1','COM2','COM3','COM4','COM5',
                     'COM6','COM7','COM8','Leave alone');
  BaudMn:  OptRec = ('110 Baud','150 Baud','300 Baud','600 Baud','1200 Baud',
                     '2400 Baud','4800 Baud','9600 Baud','Leave alone');
  ParMn:   OptRec = ('None','Odd','Even','Mark','Space',
                     'Leave alone','','','');
  DabMn:   OptRec = ('5 Bits','6 Bits','7 Bits','8 Bits','Leave alone',
                     '','','','');
  StoMn:   OptRec = ('1 Bit','2 Bits','Leave alone',
                     '','','','','','');
  MainName: String = 'Setup Main Menu.';

    {Menu function. Operands are title and list of options.}
    {Returns user selection}
  FUNCTION SuMenu (Title: String; Options: OptRec):Integer;
  VAR
    Number,i,Size: Integer;
    Tmps:          String[5];
    Indent:        String[40];
    C:             Char;
    Dummy:         Boolean;
  BEGIN
    Screen(Newl+Newl,Fromctrl);

    Size := Round( (80 - Length( Title )) / 2 );
    FOR i := 1 to Size DO
      Screen('=',Fromctrl);
    Screen(Title,Fromctrl);
    FOR i := Size + Length(Title) + 1 TO 80 DO
      Screen('=',Fromctrl);
    Screen(Newl,Fromctrl);

    Number := 0;
    Size := 12;
    WHILE
      ( Number <= 8 ) AND
      ( Length( Options[Number] ) > 0 )
    DO BEGIN                            {Find number options, }
      IF                                {Largest option}
        Length( Options [Number] ) > Size
      THEN
        Size := Length( Options[Number] );
      Number := Number + 1;
      END;

    Size := Trunc( (70 - Size) / 2 );   {Size of indent}
    Indent := '';
    FOR i := 1 TO Size DO
      Indent := Indent + ' ';
    Screen(Indent+'Key  For.....'+Newl,Fromctrl);

    FOR i := 0 TO Number-1 DO BEGIN
      Str(i+1,Tmps);
      Screen(Indent+'<'+Tmps+'>  '+Options[i]+Newl,Fromctrl);
      END;
    Screen(Newl+Indent+' Select option :',Fromctrl);

    REPEAT
      Dummy := KbdInp(C, true);  {Get a character}
    UNTIL
      ( C >= '1'  ) AND          {Is a good selection}
      ( C <= Tmps );

    Screen(C+Newl,Fromctrl);
    SuMenu := Ord(C) - Ord('1');  {Return option}
  END;

    {Generate text for main menu}
  PROCEDURE SuMain (VAR Menu: OptRec);
  VAR
    Tmps:    String[8];
  BEGIN
    Menu[0] := 'Port          (now '+ComName+' )';
    Str(Bauds[Baud], Tmps);
    Menu[1] := 'Baud rate     (now '+Tmps+' )';
    Menu[2] := 'Parity        (now '+ParNames[Parity]+' )';
    Str(DataBits+5, Tmps);
    Menu[3] := 'No. Data bits (now '+Tmps+' )';
    Str(StopBits+1, Tmps);
    Menu[4] := 'No. Stop bits (now '+Tmps+' )';
    Menu[5] := 'Continue terminal';
    Menu[6] := 'Quit terminal';
    Menu[7] := '';
    Menu[8] := '';
  END;

VAR
  Choice:    Integer;
  MainMn:    ^OptRec;
  Tmp:       String;
BEGIN
  New(MainMn);       {Create storage for main menu}

  WHILE
    true
  DO BEGIN
    SuMain(MainMn^);    {Build main menu}

    Choice := SuMenu(MainName, MainMn^);

    CASE Choice OF
      0: BEGIN { Set Com Port }
           Tmp := 'Select Port To Use (Now '+ComName+' )';
           Choice := SuMenu(Tmp, PortMn);
           IF
             Choice <> 8   {8 is Leave alone option}
           THEN BEGIN
             Tmp :=  PortMn[Choice];
             Set_Com(Tmp);  {Set port}
           END;
         END;
      1: BEGIN { Set Baud rate }
           Str(Bauds[Baud],Tmp);
           Tmp := 'Select Baud rate (now '+Tmp+' )';
           Choice := SuMenu(Tmp, BaudMn);
           IF
             Choice <> 8   {8 is Leave alone option}
           THEN BEGIN
             Str(Bauds[Choice], Tmp);
             Set_Baud(Tmp);  {Set Baud}
             END;
         END;
      2: BEGIN { Set parity }
           Tmp := 'Select Parity rate (now '+ParNames[Parity]+' )';
           Choice := SuMenu(Tmp, ParMn);
           IF
             Choice <> 5   {5 is Leave alone option}
           THEN BEGIN
             Tmp := ParMn[Choice];
             Set_Parity(Tmp);  {Set Parity}
             END;
         END;
      3: BEGIN { Set data bits }
           Str(DataBits+5, Tmp);
           Tmp := 'Select Number Data bits (now '+Tmp+' )';
           Choice := SuMenu(Tmp, DabMn);
           IF
             Choice <> 4   {4 is Leave alone option}
           THEN BEGIN
             Str(Choice + 5, Tmp);
             Set_DataBits(Tmp);  {Set Data bits}
             END;
         END;
      4: BEGIN { Set stop bits }
           Str(StopBits+1, Tmp);
           Tmp := 'Select Number Stop bits (now '+Tmp+' )';
           Choice := SuMenu(Tmp, StoMn);
           IF
             Choice <> 2   {2 is Leave alone option}
           THEN BEGIN
             Str(Choice + 1, Tmp);
             Set_StopBits(Tmp);  {Set Stop bits}
             END;
         END;
      5: BEGIN { Continue terminal }
           Screen('Setup Complete. Returning to terminal session.'+Newl,
                  Fromctrl);
           Dispose(MainMn);
           Exit;
         END;
      6: BEGIN { Quit terminal }
           CmdOut('q');
         END;
      END;
    END;
END;

  {Process menu function key - unrecognised keys are dropped}
  {#46 is passed on to command processor (is ALT-C)}
PROCEDURE MenuOut;
VAR
  c		:Char;
  AnyRead	:Boolean;
  Command       :String;
BEGIN
  AnyRead := KbdInp(c, false);	     {Get character from keyboard}
  IF
    AnyRead
  THEN
    CASE c OF
      #46: BEGIN	{ ALT-C reads in a command }
             Screen('Type command:',Fromctrl);
             Readln(Command);              {Read command}
             DropSpace(Command);           {Drop any initial spaces}
             CmdOut(Command);              {Process}
           END;
      #59:	{ F1: Help Option}
	  CmdOut('h');
      #60:	{ F2: Setup option }
	  SetUp;
      #68:	{ F10: Quit option }
	  CmdOut('q');
    END;
END;


  {Translate a character, send resulting characters to port}
  {Assumes that there is one character in string}
PROCEDURE OutXlat(Data :String);
VAR
  c,s     :Integer;
BEGIN
  WITH Out_Xlt DO BEGIN
    c := Ord(Data[1]);
    Data := '';
    FOR s := Point[c] TO Point[c+1] - 1 DO
      Data := Data + Table^[s];        {Copy translation over}
  END;
    OutStr(Data);
END;

  {Read in a character. Translate it }
FUNCTION InXlat: String;
TYPE
  Ins	= String[1];

    {Read a character from COM port.}
    {If echo to line true then send it back out}
  PROCEDURE InpStr (VAR s :Ins);
    BEGIN
    {$IFDEF INPSTR}
    Write('Entering InpStr, read');
    {$ENDIF}
    {$I-}
    Read(Com_Inp, s[1]);	{Read in Character}
    {$I+}
    IF
      IOResult = 0    	{Any read in}
    THEN BEGIN
    {$IFDEF INPSTR}
      Writeln(s[1]);
    {$ENDIF}
      s[0] := #1;       {Set one read in}
      IF
        Echol in Echo	{Echo to line}
      THEN
        OutStr(s);	{Send to port}
    END
    ELSE BEGIN
      s[0] := #0;       {Set none read}
    {$IFDEF INPSTR}
      Writeln;
    {$ENDIF}
    END;
  END;

VAR
  c,s		:Integer;
  Input 	:Ins;
  InX           :String;
BEGIN
{$IFDEF INXLAT}
  Writeln('Entering InXlat');
{$ENDIF}
  InX := '';

  InpStr(Input);	{Try and read in a character}

  IF
    Length(Input) > 0	{If any read in}
  THEN
    WITH In_Xlt DO BEGIN
      c := ord(Input[1]);
      FOR s := Point[c] TO Point[c+1] - 1 DO
	InX := InX + Table^[s];	  {Copy over}
    END;
  InXlat := InX;
END;

  {Read any characters from port, send to screen}
PROCEDURE Readin;
VAR
  Buffer	:String[255];
BEGIN
{$IFDEF READIN}
  Writeln('Entering Readin');
{$ENDIF}
  Buffer := InXlat;		{Read from port, translate}
  IF
    Length( Buffer ) > 0
  THEN
    Screen(Buffer, Fromport);	{If any read echo}
END;

  {Read characters from keyboard, If are commands process as such }
  {and send all others on to port}
PROCEDURE SwitchOut;
VAR
  AnyRead	:Boolean;
  c		:Char;
BEGIN
  AnyRead := KbdInp(c, false);		{Poll keyboard, no wait}
  IF
    AnyRead
  THEN BEGIN
    IF
      c = #0				{Is extended key code ==> command}
    THEN
      MenuOut				{Process Command/Menu}
    ELSE BEGIN
      IF				{Process normal character}
	Echos in echo			{Echo to screen set}
      THEN
	Screen(c, Fromecho);
      OutXlat(c);			{Translate and send to port}
    END;
  END;
END;

BEGIN
{  Variable initialisation section }
  ComNo    := 0;
  ComName  := 'COM1';
  Baud	   := 4;	{1200 Baud}
  Parity   := 2;	{Even}
  DataBits := 2;	{7 Bit word length}
  StopBits := 0;	{1 stop bit}
  Brk	   := 0;	{No break}

  Echo := [];		{No echo in operation}


  Init_Xlt(In_Xlt);
  Init_Xlt(Out_Xlt);	{Initialise translation tables}

  Exec_Defaults;	{Find a definition file and execute it}

  Open_com;		{Open Serial Ports}

  WHILE
    true		{Main code is infinite loop}
  DO BEGIN

    Readin;   		{Get any characters from port.}
			{Translate and send to display.}

    SwitchOut;		{Read and process characters }
			{from keyboard.}
  END;
END.
